"""Test the vasprun.xml parser."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import
# pylint: disable=invalid-name

import numpy as np
import pytest


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun(vasprun_parser):
    """Load a reference vasprun.xml and compare the result to a reference string."""

    quantity = vasprun_parser.get_quantity('occupancies')
    quantity = quantity['total']
    occ = quantity[0]
    occupancies = np.array([[[1.0, 1.0, 1.0, 1.0, 0.6667, 0.6667, 0.6667, -0.0, -0.0, -0.0]]])
    assert occ.all() == occupancies.all()


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_version(vasprun_parser):
    """Load a reference vasprun.xml and fetch the VASP version."""
    version = vasprun_parser.get_quantity('version')
    assert version == '5.4.1'


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_kpoints(vasprun_parser):
    """Load a reference vasprun.xml and test that the parsed k-points are correct."""

    kpoints = vasprun_parser.get_quantity('kpoints')
    np.testing.assert_allclose(kpoints['points'][0], np.array([0.0, 0.0, 0.0]), atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(
        kpoints['points'][-1],
        np.array([0.42857143, -0.42857143, 0.42857143]),
        atol=0.0,
        rtol=1.0e-7,
    )


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_structure(vasprun_parser):
    """Load a reference vasprun.xml and test that the parsed k-points are correct."""
    structure = vasprun_parser.get_quantity('structure')
    # Check the unit cell
    np.testing.assert_allclose(
        structure['unitcell'][0],
        np.array([5.46503124, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        structure['unitcell'][1],
        np.array([0.0, 5.46503124, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        structure['unitcell'][2],
        np.array([0.0, 0.0, 5.46503124]),
        atol=0.0,
        rtol=1.0e-7,
    )
    # Check first and last position
    np.testing.assert_allclose(
        structure['sites'][0]['position'],
        np.array([0.0, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        structure['sites'][7]['position'],
        np.array([4.09877343, 4.09877343, 1.36625781]),
        atol=0.0,
        rtol=1.0e-7,
    )


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_final_force(vasprun_parser):
    """Load a reference vasprun.xml and test that the forces are returned correctly."""
    forces = vasprun_parser.get_quantity('forces')
    forces_check = np.array(
        [
            [-0.24286901, 0.0, 0.0],
            [-0.24286901, 0.0, 0.0],
            [3.41460162, 0.0, 0.0],
            [0.44305748, 0.0, 0.0],
            [-0.73887169, 0.43727184, 0.43727184],
            [-0.94708885, -0.85011586, 0.85011586],
            [-0.94708885, 0.85011586, -0.85011586],
            [-0.73887169, -0.43727184, -0.43727184],
        ]
    )
    # Check first, third and last position
    np.testing.assert_allclose(forces[0], forces_check[0], atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(forces[2], forces_check[2], atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(forces[7], forces_check[7], atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_final_stress(vasprun_parser):
    """Load a reference vasprun.xml and test that the stress are returned correctly."""
    stress = vasprun_parser.get_quantity('stress')
    stress_check = np.array(
        [
            [-0.38703740, 0.00000000, 0.00000000],
            [0.00000000, 12.52362644, -25.93894358],
            [0.00000000, -25.93894358, 12.52362644],
        ]
    )

    # Check entries
    np.testing.assert_allclose(stress[0], stress_check[0], atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(stress[1], stress_check[1], atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(stress[2], stress_check[2], atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(['vasprun_parser'], [('dielectric',)], indirect=True)
def test_parse_vasprun_dielectrics(vasprun_parser):
    """Load a reference vasprun.xml and test that the dielectrics are returned correctly."""
    dielectrics = vasprun_parser.get_quantity('dielectrics')
    imag = dielectrics['idiel']
    real = dielectrics['rdiel']
    energy = dielectrics['ediel']
    # Test shape of arrays
    assert imag.shape == (1000, 6)
    assert real.shape == (1000, 6)
    assert energy.shape == (1000,)
    # Test a few entries
    np.testing.assert_allclose(imag[0], np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0]), atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(
        imag[500],
        np.array([0.0933, 0.0924, 0.0924, 0.0, 0.0082, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )

    np.testing.assert_allclose(
        imag[999],
        np.array([0.0035, 0.0035, 0.0035, 0.0, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        real[0],
        np.array([12.0757, 11.4969, 11.4969, 0.0, 0.6477, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        real[500],
        np.array([-0.5237, -0.5366, -0.5366, 0.0, 0.0134, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        real[999],
        np.array([6.57100000e-01, 6.55100000e-01, 6.55100000e-01, 0.0, -1.00000000e-04, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    assert energy[500] == pytest.approx(10.2933)


@pytest.mark.parametrize(['vasprun_parser'], [('disp_details',)], indirect=True)
def test_parse_vasprun_epsilon(vasprun_parser):
    """Load a reference vasprun.xml and test that epsilon is returned correctly."""
    result = vasprun_parser.get_quantity('dielectrics')
    epsilon = result['epsilon']
    epsilon_ion = result['epsilon_ion']
    # Test shape of arrays
    assert epsilon.shape == (3, 3)
    assert epsilon_ion.shape == (3, 3)
    # Test a few entries
    test = np.array([[13.05544887, -0.0, 0.0], [-0.0, 13.05544887, -0.0], [0.0, 0.0, 13.05544887]])
    np.testing.assert_allclose(epsilon, test, atol=0.0, rtol=1.0e-7)
    test = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
    np.testing.assert_allclose(epsilon_ion, test, atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(['vasprun_parser'], [('localfield',)], indirect=True)
def test_parse_vasprun_born(vasprun_parser):
    """Load a reference vasprun.xml and test that the Born effective charges are
    returned correctly."""
    born = vasprun_parser.get_quantity('born_charges')
    # Test shape of array
    assert born.shape == (8, 3, 3)
    # Test a few entries
    np.testing.assert_allclose(born[0][0], np.array([6.37225000e-03, 0.0, 0.0]), atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(
        born[0][-1],
        np.array([-4.21760000e-04, -2.19570210e-01, 3.20709600e-02]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        born[4][0],
        np.array([1.68565200e-01, -2.92058000e-02, -2.92058000e-02]),
        atol=0.0,
        rtol=1.0e-7,
    )


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_dos(vasprun_parser):
    """Load a reference vasprun.xml and test that the density of states are
    returned correctly."""
    result = vasprun_parser.get_quantity('dos')
    dos = result['tdos']
    energy = result['energy']
    # Test shape of array
    assert dos.shape == (301,)
    assert energy.shape == (301,)
    # Test a few entries
    assert dos[150] == pytest.approx(4.1296)
    assert energy[150] == pytest.approx(2.3373)


@pytest.mark.parametrize(['vasprun_parser'], [('spin',)], indirect=True)
def test_parse_vasprun_dos_spin(vasprun_parser):
    """Load a reference vasprun.xml and test that the spin decomposed
    density of states is returned correctly."""
    result = vasprun_parser.get_quantity('dos')
    dos = result['tdos']
    # Test shape of array
    assert dos.shape == (
        2,
        1000,
    )
    # Test a few entries
    assert dos[0, 500] == pytest.approx(0.9839)
    assert dos[1, 500] == pytest.approx(0.9844)


@pytest.mark.parametrize(['vasprun_parser'], [('partial',)], indirect=True)
def test_parse_vasprun_pdos(vasprun_parser):
    """Load a reference vasprun.xml and test that the projected
    density of states is returned correctly."""
    result = vasprun_parser.get_quantity('dos')
    dos = result['pdos']
    energy = result['energy']
    # Test shape of array
    assert dos.shape == (8, 1000, 9)
    assert energy.shape == (1000,)
    # Test a few entries
    np.testing.assert_allclose(
        dos[3, 500],
        np.array([0.0770, 0.0146, 0.0109, 0.0155, 0.0, 0.0, 0.0, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        dos[7, 500],
        np.array([0.0747, 0.0121, 0.0092, 0.0116, 0.0, 0.0, 0.0, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    assert energy[500] == pytest.approx(0.01)


@pytest.mark.parametrize(['vasprun_parser'], [('partial',)], indirect=True)
def test_parse_vasprun_projectors(vasprun_parser):
    """Load a reference vasprun.xml and test that the state projectors are
    returned correctly."""
    proj = vasprun_parser.get_quantity('projectors')
    # Test shape of array
    assert proj.shape == (8, 64, 21, 9)
    # Test a few entries
    np.testing.assert_allclose(
        proj[0, 0, 5],
        np.array([0.0, 0.012, 0.0123, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        proj[7, 0, 5],
        np.array([0.1909, 0.0001, 0.0001, 0.0001, 0.0, 0.0, 0.0, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )
    np.testing.assert_allclose(
        proj[4, 3, 5],
        np.array([0.2033, 0.0001, 0.0001, 0.0001, 0.0, 0.0, 0.0, 0.0, 0.0]),
        atol=0.0,
        rtol=1.0e-7,
    )


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_eigenvalues_occupancies(vasprun_parser):
    """Load a reference vasprun.xml and test that the eigenvalues are
    returned correctly."""
    eigen = vasprun_parser.get_quantity('eigenvalues')
    eigen = eigen['total']
    occ = vasprun_parser.get_quantity('occupancies')
    occ = occ['total']
    # Test shape of array
    assert eigen.shape == (64, 21)
    assert occ.shape == (64, 21)
    # Test a few entries
    assert eigen[0, 0] == pytest.approx(-6.2348)
    assert eigen[0, 15] == pytest.approx(5.8956)
    assert eigen[6, 4] == pytest.approx(-1.7424)
    assert occ[0, 0] == pytest.approx(1.0)
    assert occ[0, 15] == pytest.approx(0.6949)
    assert occ[6, 4] == pytest.approx(1.0)


@pytest.mark.parametrize(['vasprun_parser'], [('spin',)], indirect=True)
def test_parse_vasprun_eigenocc_spin_result(vasprun_parser):
    """Load a reference vasprun.xml and test that the spin decomposed eigenvalues
    are returned correctly."""
    eigen = vasprun_parser.get_quantity('eigenvalues')
    occ = vasprun_parser.get_quantity('occupancies')
    # Test shape of array
    assert eigen['up'].shape == (64, 25)
    assert occ['up'].shape == (64, 25)
    # Test a few entries
    assert eigen['up'][0, 0] == pytest.approx(-6.2363)
    assert eigen['up'][0, 15] == pytest.approx(5.8939)
    assert eigen['up'][6, 4] == pytest.approx(-1.7438)
    assert eigen['down'][0, 0] == pytest.approx(-6.2357)
    assert eigen['down'][0, 15] == pytest.approx(5.8946)
    assert eigen['down'][6, 4] == pytest.approx(-1.7432)
    assert occ['up'][0, 0] == pytest.approx(1.0)
    assert occ['up'][0, 15] == pytest.approx(0.6955)
    assert occ['up'][6, 4] == pytest.approx(1.0)
    assert occ['down'][0, 0] == pytest.approx(1.0)
    assert occ['down'][0, 15] == pytest.approx(0.6938)
    assert occ['down'][6, 4] == pytest.approx(1.0)


@pytest.mark.parametrize(['vasprun_parser'], [('basic',)], indirect=True)
def test_parse_vasprun_toten(vasprun_parser):
    """Load a reference vasprun.xml and test that one of the  total energies
    is returned correctly."""
    result = vasprun_parser.get_quantity('energies')
    assert set(result.keys()) == set(['energy_extrapolated', 'energy_extrapolated_electronic', 'electronic_steps'])
    energies = result['energy_extrapolated']
    test_array = np.array([-42.91113621])
    np.testing.assert_allclose(test_array, energies, atol=0.0, rtol=1.0e-7)
    # Test number of entries
    assert energies.shape == (1,)
    # Electronic steps should be one
    test_array = np.array([1])
    np.testing.assert_allclose(test_array, result['electronic_steps'], atol=0.0, rtol=1.0e-7)
    # Testing on VASP 5, where the extrapolated energy should be the following due to a bug
    test_array = np.array([-0.00236711])
    with np.testing.assert_raises(AssertionError):
        np.testing.assert_allclose(test_array, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)
    # Instead we correct and it should be
    test_array = np.array([-42.911136])
    np.testing.assert_allclose(test_array, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(
    ['vasprun_parser'],
    [
        (
            [
                'basic',
                'vasprun.xml',
                {'energy_type': ['energy_free', 'energy_no_entropy']},
            ],
        )
    ],
    indirect=True,
)
def test_toten_multiple(vasprun_parser):
    """Load a reference vasprun.xml and test that multiple total energies
    are returned properly."""
    result = vasprun_parser.get_quantity('energies')
    assert set(result.keys()) == set(
        [
            'electronic_steps',
            'energy_free_electronic',
            'energy_free',
            'energy_no_entropy',
            'energy_no_entropy_electronic',
        ]
    )
    test_array = np.array([-42.91231976])
    np.testing.assert_allclose(test_array, result['energy_free_electronic'], atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(test_array, result['energy_free'], atol=0.0, rtol=1.0e-7)
    test_array = np.array([-42.90995265])
    np.testing.assert_allclose(test_array, result['energy_no_entropy_electronic'], atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(test_array, result['energy_no_entropy'], atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(
    ['vasprun_parser'],
    [(['basic', 'vasprun.xml', {'electronic_step_energies': True}],)],
    indirect=True,
)
def test_parse_vasprun_toten_electronic(vasprun_parser):
    """Load a reference vasprun.xml and test that the total energies
    are returned correctly for the electronic steps."""
    result = vasprun_parser.get_quantity('energies')
    # Test that the default arrays are present
    assert set(result.keys()) == set(['energy_extrapolated', 'energy_extrapolated_electronic', 'electronic_steps'])
    energies = result['energy_extrapolated_electronic']
    test_array = np.array([-42.91113666, -42.91113621])
    np.testing.assert_allclose(test_array, energies, atol=0.0, rtol=1.0e-7)
    # Test number of entries
    assert energies.shape == (2,)
    # Electronic steps should be two
    test_array = np.array([2])
    np.testing.assert_allclose(test_array, result['electronic_steps'], atol=0.0, rtol=1.0e-7)
    # Testing on VASP 5, where the extrapolated energy should be the following due to a bug
    test_array = np.array([-0.00236711])
    with np.testing.assert_raises(AssertionError):
        np.testing.assert_allclose(test_array, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)
    # Instead we correct and it should be
    test_array = np.array([-42.911136])
    np.testing.assert_allclose(test_array, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(['vasprun_parser'], [('relax',)], indirect=True)
def test_parse_vasprun_toten_relax(vasprun_parser):
    """Load a reference vasprun.xml and check that the total energies are
    returned correctly for relaxation runs."""
    result = vasprun_parser.get_quantity('energies')
    assert set(result.keys()) == set(['energy_extrapolated', 'energy_extrapolated_electronic', 'electronic_steps'])
    energies = result['energy_extrapolated_electronic']
    test_array = np.array(
        [
            -42.91113348,
            -43.27757545,
            -43.36648855,
            -43.37734069,
            -43.38062479,
            -43.38334165,
            -43.38753003,
            -43.38708193,
            -43.38641449,
            -43.38701639,
            -43.38699488,
            -43.38773717,
            -43.38988315,
            -43.3898822,
            -43.39011239,
            -43.39020751,
            -43.39034244,
            -43.39044584,
            -43.39087657,
        ]
    )
    # Test energies
    np.testing.assert_allclose(test_array, energies, atol=0.0, rtol=1.0e-7)
    # Test number of entries
    assert energies.shape == test_array.shape
    # Electronic steps should be entries times one
    np.testing.assert_allclose(np.ones(19, dtype=int), result['electronic_steps'], atol=0.0, rtol=1.0e-7)
    # Testing on VASP 5, where the extrapolated energy should be the following due to a bug
    test_array = np.array(
        [
            -0.00236637,
            -0.00048614,
            -0.00047201,
            -0.00043261,
            -0.00041668,
            -0.00042584,
            -0.00043637,
            -0.00042806,
            -0.00042762,
            -0.00043875,
            -0.00042731,
            -0.00042705,
            -0.00043064,
            -0.00043051,
            -0.00043161,
            -0.00043078,
            -0.00043053,
            -0.00043149,
            -0.00043417,
        ]
    )
    with np.testing.assert_raises(AssertionError):
        np.testing.assert_allclose(test_array, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)
    # Instead we correct and it should be
    test_array = np.array(
        [
            -42.911133,
            -43.277575,
            -43.366489,
            -43.377341,
            -43.380625,
            -43.383342,
            -43.38753,
            -43.387082,
            -43.386414,
            -43.387016,
            -43.386995,
            -43.387737,
            -43.389883,
            -43.389882,
            -43.390112,
            -43.390208,
            -43.390342,
            -43.390446,
            -43.390877,
        ]
    )
    np.testing.assert_allclose(test_array, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(
    ['vasprun_parser'],
    [(['relax', 'vasprun.xml', {'electronic_step_energies': True}],)],
    indirect=True,
)
def test_parse_vasprun_toten_relax_electronic(vasprun_parser):
    """Load a reference vasprun.xml and check that the total energies
    are returned correctly for both the electronic and ionic steps."""
    result = vasprun_parser.get_quantity('energies')
    assert set(result.keys()) == set(['energy_extrapolated', 'energy_extrapolated_electronic', 'electronic_steps'])
    energies = result['energy_extrapolated_electronic']
    test_array_energies = [
        np.array(
            [
                163.37398579,
                14.26925896,
                -23.05190509,
                -34.91615104,
                -40.20080347,
                -42.18390876,
                -42.97469852,
                -43.31556073,
                -43.60169068,
                -43.61723125,
                -43.61871511,
                -43.61879751,
                -43.12548175,
                -42.90647187,
                -42.91031846,
                -42.91099027,
                -42.91111107,
                -42.91113348,
            ]
        ),
        np.array(
            [
                -43.34236449,
                -43.31102002,
                -43.27768275,
                -43.27791002,
                -43.27761357,
                -43.27757545,
            ]
        ),
        np.array(
            [
                -43.40320524,
                -43.38084022,
                -43.36835045,
                -43.36666248,
                -43.36666583,
                -43.36649036,
                -43.36648855,
            ]
        ),
        np.array([-43.37749056, -43.37749102, -43.37734414, -43.37734069]),
        np.array([-43.38117265, -43.38082881, -43.38063293, -43.38062479]),
        np.array([-43.38337336, -43.38334165]),
        np.array([-43.38778922, -43.38766017, -43.38752953, -43.38753003]),
        np.array([-43.38714489, -43.38708193]),
        np.array([-43.38640951, -43.38641449]),
        np.array([-43.3874799, -43.3871553, -43.38701949, -43.38701639]),
        np.array([-43.38790942, -43.38727062, -43.38700335, -43.38699488]),
        np.array([-43.38774394, -43.38773717]),
        np.array([-43.38984942, -43.3899134, -43.38988315]),
        np.array([-43.38988117, -43.3898822]),
        np.array([-43.39032165, -43.39017866, -43.39011239]),
        np.array([-43.39021044, -43.39020751]),
        np.array([-43.39034135, -43.39034244]),
        np.array([-43.39044466, -43.39044584]),
        np.array([-43.39084354, -43.39088709, -43.39087657]),
    ]
    test_array_steps = np.array([18, 6, 7, 4, 4, 2, 4, 2, 2, 4, 4, 2, 3, 2, 3, 2, 2, 2, 3])
    # Build a flattened array (not using flatten from NumPy as the content is staggered) and
    # test number of electronic steps per ionic step
    test_array_energies_flattened = np.array([])
    for ionic_step in test_array_energies:
        test_array_energies_flattened = np.append(test_array_energies_flattened, ionic_step)
    assert energies.shape == test_array_energies_flattened.shape
    np.testing.assert_allclose(test_array_energies_flattened, energies, atol=0.0, rtol=1.0e-7)
    np.testing.assert_allclose(test_array_steps, result['electronic_steps'], atol=0.0, rtol=1.0e-7)
    test_array_energies = np.array(
        [
            -0.00236637,
            -0.00048614,
            -0.00047201,
            -0.00043261,
            -0.00041668,
            -0.00042584,
            -0.00043637,
            -0.00042806,
            -0.00042762,
            -0.00043875,
            -0.00042731,
            -0.00042705,
            -0.00043064,
            -0.00043051,
            -0.00043161,
            -0.00043078,
            -0.00043053,
            -0.00043149,
            -0.00043417,
        ]
    )
    # Testing on VASP 5, where the extrapolated energy should be the following due to a bug
    with np.testing.assert_raises(AssertionError):
        np.testing.assert_allclose(test_array_energies, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)
    # Instead we correct and it should be
    test_array_energies = np.array(
        [
            -42.911133,
            -43.277575,
            -43.366489,
            -43.377341,
            -43.380625,
            -43.383342,
            -43.38753,
            -43.387082,
            -43.386414,
            -43.387016,
            -43.386995,
            -43.387737,
            -43.389883,
            -43.389882,
            -43.390112,
            -43.390208,
            -43.390342,
            -43.390446,
            -43.390877,
        ]
    )
    np.testing.assert_allclose(test_array_energies, result['energy_extrapolated'], atol=0.0, rtol=1.0e-7)


@pytest.mark.parametrize(['vasprun_parser'], [('disp',)], indirect=True)
def test_parse_vasprun_hessian(vasprun_parser):
    """Load a reference vasprun.xml and check that the Hessian matrix
    are returned correctly."""
    hessian = vasprun_parser.get_quantity('hessian')
    # Test shape
    assert hessian.shape == (24, 24)
    # Test a few entries
    assert np.allclose(
        hessian[0],
        np.array(
            [
                -4.63550410e-01,
                0.00000000e00,
                0.00000000e00,
                -5.91774100e-02,
                0.00000000e00,
                0.00000000e00,
                3.09711000e-02,
                0.00000000e00,
                0.00000000e00,
                3.20435400e-02,
                0.00000000e00,
                0.00000000e00,
                1.15129840e-01,
                -8.16138200e-02,
                8.17234700e-02,
                1.14879520e-01,
                8.11324800e-02,
                8.27409500e-02,
                1.14879520e-01,
                -8.11324800e-02,
                -8.27409500e-02,
                1.15129840e-01,
                8.16138200e-02,
                -8.17234700e-02,
            ]
        ),
    )
    assert np.allclose(
        hessian[-2],
        np.array(
            [
                8.16138200e-02,
                1.15195590e-01,
                -8.38411100e-02,
                -8.17234700e-02,
                1.14875090e-01,
                -8.53388100e-02,
                3.46686900e-02,
                7.00672700e-02,
                2.54288300e-02,
                -8.26222700e-02,
                1.16185510e-01,
                7.95575600e-02,
                -3.05970000e-04,
                3.16827300e-02,
                2.86379000e-03,
                5.42080000e-04,
                3.27613500e-02,
                1.12576000e-03,
                -1.34305000e-03,
                -5.86811100e-02,
                2.83374000e-03,
                4.91688400e-02,
                -4.22101090e-01,
                5.73736900e-02,
            ]
        ),
    )


@pytest.mark.parametrize(['vasprun_parser'], [('disp',)], indirect=True)
def test_parse_vasprun_dynmat(vasprun_parser):
    """Load a reference vasprun.xml and check that the dynamical eigenvectors and eigenvalues
    are returned correctly."""
    result = vasprun_parser.get_quantity('dynmat')
    dynvec = result['dynvec']
    dyneig = result['dyneig']
    # test shape
    assert dynvec.shape == (24, 24)
    assert dyneig.shape == (24,)
    # test a few entries
    assert np.allclose(
        dynvec[0],
        np.array(
            [
                7.28517310e-17,
                7.25431601e-02,
                -4.51957676e-02,
                1.15412776e-16,
                4.51957676e-02,
                -7.25431601e-02,
                -1.37347223e-16,
                5.16257351e-01,
                -5.16257351e-01,
                8.16789156e-17,
                8.95098005e-02,
                -8.95098005e-02,
                -4.43838008e-17,
                -6.38031134e-02,
                6.38031134e-02,
                -1.80132830e-01,
                -2.97969516e-01,
                2.97969516e-01,
                1.80132830e-01,
                -2.97969516e-01,
                2.97969516e-01,
                -2.09989969e-16,
                -6.38031134e-02,
                6.38031134e-02,
            ]
        ),
    )
    assert np.allclose(
        dynvec[4],
        np.array(
            [
                -5.29825122e-13,
                -2.41759046e-01,
                -3.28913434e-01,
                -5.30734671e-13,
                -3.28913434e-01,
                -2.41759046e-01,
                3.26325910e-13,
                -3.80807441e-02,
                -3.80807441e-02,
                -9.22956103e-13,
                -2.99868012e-01,
                -2.99868012e-01,
                1.64418993e-01,
                1.81002749e-01,
                1.81002749e-01,
                3.11984195e-13,
                2.73349550e-01,
                2.73349550e-01,
                2.59853610e-13,
                2.73349550e-01,
                2.73349550e-01,
                -1.64418993e-01,
                1.81002749e-01,
                1.81002749e-01,
            ]
        ),
    )
    assert dyneig[0] == pytest.approx(-1.36621537e00)
    assert dyneig[4] == pytest.approx(-8.48939361e-01)


@pytest.mark.parametrize(['vasprun_parser'], [('spin',)], indirect=True)
def test_band_properties(aiida_profile, vasprun_parser):
    """Load a reference vasprun.xml and check that key properties of the electric structure
    are returned correctly."""
    data = vasprun_parser.get_quantity('band_properties')
    assert data['cbm'] == pytest.approx(6.5536)
    assert data['vbm'] == pytest.approx(6.5105)
    assert data['is_direct_gap'] is False
    assert data['band_gap'] == pytest.approx(0.04310, rel=1e-3)



"""
The ``CHGCAR`` parser interface.

Contains the parsing interfaces to ``parsevasp`` used to parse ``CHGCAR`` content.
"""

# pylint: disable=abstract-method
from typing import TextIO

from parsevasp.chgcar import Chgcar

from aiida_vasp.parsers.content_parsers.base import BaseFileParser


class ChgcarParser(BaseFileParser):
    """The parser interface that enables parsing of ``CHGCAR`` content.

    The parser is triggered by using the ``charge_density`` and/or ``magnetization_density`` quantity key.

    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['charge_density']}

    PARSABLE_QUANTITIES = {
        'charge_density': {'inputs': [], 'name': 'charge_density', 'prerequisites': []},
        'magnetization_density': {'inputs': [], 'name': 'magnetization_density', 'prerequisites': []},
    }

    def _init_from_handler(self, handler: TextIO) -> None:
        """Initialize a ``parsevasp`` object of ``Chgcar`` using a file like handler.

        :param handler: A file like object that provides the necessary ``CHGCAR`` content to be parsed.
        :type handler: file-like object
        """

        try:
            self._content_parser = Chgcar(file_handler=handler, logger=self._logger)
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    @property
    def charge_density(self):
        """Return the charge density.

        :returns: A NumPy array containing the charge density in the unit cell in C order.
        :rtype: ndarray
        """

        return self._content_parser.charge_density

    @property
    def magnetization_density(self):
        """Return the magnetization density.

        :returns: If collinear spin calculations have been performed, a NumPy array containing
                  the magnetization density in the unit cell in C order is returned. If however
                  a non-collinear spin calculation have been performed, a dictionary is returned
                  with keys `x`, `y` and `z`, each containing the same NumPy array, but for each
                  direction.
        :rtype: dict or ndarray
        """
        return self._content_parser.magnetization_density



"""
Test submitting a VaspWorkChain.

This does not seem to work, for `submit` the daemon will not pick up the workchain
and `run` just seems to get stuck after a while.
"""

# pylint: disable=unused-import,wildcard-import,unused-wildcard-import,unused-argument,redefined-outer-name, import-outside-toplevel
from __future__ import print_function

import numpy as np
import pytest
from aiida import orm
from aiida.cmdline.utils.common import get_calcjob_report, get_workchain_report
from aiida.common.exceptions import InputValidationError
from aiida.common.extendeddicts import AttributeDict
from aiida.engine import run
from aiida.plugins import WorkflowFactory
from aiida.plugins.factories import DataFactory


@pytest.mark.parametrize(['vasp_structure', 'vasp_kpoints'], [('str', 'mesh')], indirect=True)
def test_vasp_wc(run_vasp_process):
    """Test submitting only, not correctness, with mocked vasp code."""
    results, node = run_vasp_process(process_type='workchain')
    assert node.exit_status == 0
    assert 'retrieved' in results
    assert 'misc' in results
    assert 'remote_folder' in results
    misc = results['misc'].get_dict()
    assert np.amax(np.linalg.norm(misc['stress'], axis=1)) == pytest.approx(22.8499295)
    assert misc['total_energies']['energy_extrapolated'] == pytest.approx(-14.16209692)

    assert 'stdout' in node.called[0].inputs.monitors
    assert 'loop_time' in node.called[0].inputs.monitors


def si_structure():
    """
    Setup a silicon structure in a displaced FCC setting
    """
    structure_data = DataFactory('core.structure')
    alat = 3.9
    lattice = np.array([[0.5, 0.5, 0], [0, 0.5, 0.5], [0.5, 0, 0.5]]) * alat
    structure = structure_data(cell=lattice)
    positions = [[0.1, 0.0, 0.0]]
    for pos_direct in positions:
        pos_cartesian = np.dot(pos_direct, lattice)
        structure.append_atom(position=pos_cartesian, symbols='Si')
    return structure


# TEST INPUT FOR AUTOMATIC correction of NELM
# calculation should finish in the second run where the calculation
INCAR_ELEC_CONV = {
    'encut': 240,
    'ismear': 0,
    'sigma': 0.1,
    'ediff': 1e-9,
    'nelm': 7,
    'ibrion': -1,
    'potim': 0.01,
    'nsw': -1,
    'isif': 3,
    # 'ediffg': -0.01
}

INCAR_IONIC_CONV = {
    'encut': 240,
    'ismear': 0,
    'sigma': 0.1,
    'ediff': 1e-9,
    'nelm': 15,
    'ibrion': 1,
    'potim': 0.1,
    'nsw': 5,
    'isif': 3,
}

# Parameters for test handling unfinished VASP. The first iteration was killed manually.
INCAR_IONIC_UNFINISHED = {
    'encut': 500,
    'ismear': 0,
    'isym': 0,
    'sigma': 0.1,
    'ediff': 1e-9,
    'nelm': 15,
    'ibrion': 1,
    'potim': 0.1,
    'nsw': 20,
    'isif': 3,
}


def setup_vasp_workchain(structure, incar, nkpts, potcar_family_name, potcar_mapping, code=None):
    """
    Setup the inputs for a VaspWorkChain.
    """

    inputs = AttributeDict()

    inputs.structure = structure
    inputs.parameters = orm.Dict(dict={'incar': incar})

    kpoints = orm.KpointsData()
    kpoints.set_kpoints_mesh((nkpts, nkpts, nkpts))
    inputs.kpoints = kpoints

    inputs.potential_family = orm.Str(potcar_family_name)
    inputs.potential_mapping = orm.Dict(dict=potcar_mapping)
    inputs.options = orm.Dict(
        dict={
            'withmpi': False,
            'queue_name': 'None',
            'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1},
            'max_wallclock_seconds': 3600,
        },
    )
    inputs.settings = orm.Dict(dict={'parser_settings': {'add_structure': True}})

    # If code is not passed, use the mock code
    if code is None:
        mock = orm.load_code('mock-vasp@localhost')
        inputs.code = mock
    else:
        inputs.code = code
    return inputs


def test_vasp_incar_case(aiida_profile):
    """Test catching inconsistent incar key cases"""
    workchain = WorkflowFactory('vasp.vasp')
    builder = workchain.get_builder()

    with pytest.warns(UserWarning, match="Key 'ALGO' converted to 'algo'"):
        builder.parameters = orm.Dict({'incar': {'ALGO': 21}})

    parameters = orm.Dict({'incar': {'ALGO': 21}})
    parameters.store()

    with pytest.raises(InputValidationError, match='Case inconsistency found in'):
        builder.parameters = parameters


def test_vasp_wc_nelm(aiida_profile, upload_potcar, potcar_family_name, potcar_mapping, mock_vasp_strict):
    """Test with mocked vasp code for handling electronic convergence issues"""

    workchain = WorkflowFactory('vasp.vasp')

    mock_vasp_strict.store()

    # create_authinfo(computer=mock_vasp_strict.computer, store=True)

    inputs = setup_vasp_workchain(si_structure(), INCAR_ELEC_CONV, 8, potcar_family_name, potcar_mapping)
    inputs.verbose = orm.Bool(True)
    results, node = run.get_node(workchain, **inputs)

    called_nodes = list(node.called)
    called_nodes.sort(key=lambda x: x.ctime)

    print(get_workchain_report(node, 'DEBUG'))
    for child in called_nodes:
        print(get_calcjob_report(child))

    child = called_nodes[0]
    print(child.base.repository.get_object_content('INCAR'))
    print(child.base.repository.get_object_content('POSCAR'))
    print(child.base.repository.get_object_content('KPOINTS'))
    print(child.outputs.retrieved.base.repository.get_object_content('vasp_output'))
    print(child.outputs.retrieved.base.repository.list_object_names())
    print(child.outputs.misc.get_dict())
    print(child.exit_status)

    child = called_nodes[1]
    print(child.base.repository.get_object_content('INCAR'))
    print(child.base.repository.get_object_content('POSCAR'))
    print(child.base.repository.get_object_content('KPOINTS'))
    print(child.outputs.retrieved.base.repository.get_object_content('vasp_output'))
    print(child.outputs.retrieved.base.repository.list_object_names())
    print(child.outputs.misc.get_dict())
    print(child.exit_status)

    assert node.exit_status == 0
    assert 'retrieved' in results
    assert 'misc' in results
    assert 'remote_folder' in results

    assert results['misc']['total_energies']['energy_extrapolated'] == pytest.approx(-4.82467802)

    # Sort the called nodes by creation time
    called_nodes = list(node.called)
    called_nodes.sort(key=lambda x: x.ctime)

    assert called_nodes[0].exit_status == 701
    assert called_nodes[1].exit_status == 0


@pytest.mark.parametrize(
    'incar,nkpts,exit_codes',
    [[INCAR_IONIC_CONV, 8, [702, 0]], [INCAR_IONIC_UNFINISHED, 16, [700, 0]]],
)
def test_vasp_wc_ionic_continue(
    aiida_profile, upload_potcar, potcar_family_name, potcar_mapping, mock_vasp_strict, incar, nkpts, exit_codes
):
    """Test with mocked vasp code for handling ionic convergence issues"""
    workchain = WorkflowFactory('vasp.vasp')

    mock_vasp_strict.store()
    # create_authinfo(computer=mock_vasp_strict.computer, store=True)

    inputs = setup_vasp_workchain(si_structure(), incar, nkpts, potcar_family_name, potcar_mapping)
    inputs.verbose = orm.Bool(True)
    # The test calculation contain NELM breaches during the relaxation - set to ignore it.
    inputs.handler_overrides = orm.Dict(dict={'ignore_nelm_breach_relax': True})
    results, node = run.get_node(workchain, **inputs)

    assert node.exit_status == 0
    assert 'retrieved' in results
    assert 'misc' in results
    assert 'remote_folder' in results

    assert results['misc']['run_status']['ionic_converged']

    # Sort the called nodes by creation time
    called_nodes = list(node.called)
    called_nodes.sort(key=lambda x: x.ctime)

    # Check the child status - here the first calculation is not finished but the second one is
    for idx, code in enumerate(exit_codes):
        assert called_nodes[idx].exit_status == code


@pytest.mark.skip(reason='This test is not working yet')
def test_vasp_wc_ionic_magmom_carry(aiida_profile, upload_potcar, potcar_family_name, potcar_mapping, mock_vasp_strict):
    """Test with mocked vasp code for handling ionic convergence issues"""

    workchain = WorkflowFactory('vasp.vasp')

    mock_vasp_strict.store()
    # create_authinfo(computer=mock_vasp_strict.computer, store=True)

    incar = dict(INCAR_IONIC_CONV)
    incar['ispin'] = 2
    incar['lorbit'] = 10
    incar['nupdown'] = 2
    inputs = setup_vasp_workchain(si_structure(), incar, 8, potcar_family_name, potcar_mapping)
    inputs.verbose = orm.Bool(True)

    # The test calculation contain NELM breaches during the relaxation - set to ignore it.
    inputs.handler_overrides = orm.Dict(dict={'ignore_nelm_breach_relax': True})
    inputs.settings = orm.Dict(
        dict={
            'parser_settings': {
                'add_structure': True,
                'add_site_magnetization': True,
            }
        },
    )
    inputs.max_iterations = orm.Int(2)

    _, node = run.get_node(workchain, **inputs)
    assert node.exit_status == 0

    called_nodes = list(node.called)
    called_nodes.sort(key=lambda x: x.ctime)
    # Check that the second node takes the magnetization of the first node
    assert called_nodes[1].inputs.parameters['magmom'] == [0.646]



"""Test the CHGCAR parser."""
# pylint: disable=unused-import,redefined-outer-name,unused-argument,unused-wildcard-import,wildcard-import

import numpy as np
import pytest

compare_charge_density = np.array(
    [
        [
            [0.09329446, 0.18658892, 0.27988338],
            [0.37317784, 0.4664723, 0.55976676],
            [0.65306122, 0.74635569, 0.83965015],
            [0.93294461, 1.02623907, 1.11953353],
        ],
        [
            [1.21282799, 1.30612245, 1.39941691],
            [1.49271137, 1.58600583, 1.67930029],
            [1.77259475, 1.86588921, 1.95918367],
            [2.05247813, 2.14577259, 2.23906706],
        ],
        [
            [2.33236152, 2.42565598, 2.51895044],
            [2.6122449, 2.70553936, 2.79883382],
            [2.89212828, 2.98542274, 3.0787172],
            [3.17201166, 3.26530612, 3.35860058],
        ],
        [
            [3.45189504, 3.5451895, 3.63848397],
            [3.73177843, 3.82507289, 3.91836735],
            [4.01166181, 4.10495627, 4.19825073],
            [4.29154519, 4.38483965, 4.47813411],
        ],
        [
            [4.57142857, 4.66472303, 4.75801749],
            [4.85131195, 4.94460641, 5.03790087],
            [5.13119534, 5.2244898, 5.31778426],
            [5.41107872, 5.50437318, 5.59766764],
        ],
    ]
)


@pytest.mark.parametrize(['chgcar_parser'], [('chgcar',)], indirect=True)
def test_parse_charge(chgcar_parser):
    """Load a reference CHGCAR parser.

    We check that the it parses and provides the correct content for the charge density.

    """
    result = chgcar_parser.get_quantity('charge_density')
    assert result is not None
    assert np.allclose(result, compare_charge_density)


@pytest.mark.parametrize(
    ['chgcar_parser'],
    [
        (
            [
                'chgcar',
                'CHGCAR.spin',
                {'quantities_to_parse': ['charge_density', 'magnetization_density']},
            ],
        )
    ],
    indirect=True,
)
def test_parse_magnetization(chgcar_parser):
    """Load a reference CHGCAR parser.

    We check that the it parses and provides the correct content for the magnetization density.

    """
    result = chgcar_parser.get_quantity('charge_density')
    assert result is not None
    assert np.allclose(result, compare_charge_density)
    # Magnetization density if turned off by default, enable it
    result = chgcar_parser.get_quantity('magnetization_density')
    assert result is not None
    assert np.allclose(result, compare_charge_density)


@pytest.mark.parametrize(
    ['chgcar_parser'],
    [
        (
            [
                'chgcar',
                'CHGCAR.ncl',
                {'quantities_to_parse': ['charge_density', 'magnetization_density']},
            ],
        )
    ],
    indirect=True,
)
def test_parse_magnetization_ncl(chgcar_parser):
    """Load a reference CHGCAR parser.

    We check that the it parses and provides the correct content for the magnetization density.

    """
    result = chgcar_parser.get_quantity('charge_density')
    assert result is not None
    assert np.allclose(result, compare_charge_density)
    result = chgcar_parser.get_quantity('magnetization_density')
    assert result is not None
    assert set(['x', 'y', 'z']) == set(result.keys())
    for item in result.values():
        assert np.allclose(item, compare_charge_density)



"""
VASP NEB workchain.

Contains the VaspNEBWorkChain class definition which uses the BaseRestartWorkChain.
"""

from __future__ import annotations

from typing import Any

import numpy as np
from aiida import __version__ as aiida_version
from aiida import orm
from aiida.common.exceptions import InputValidationError, NotExistent
from aiida.common.extendeddicts import AttributeDict
from aiida.common.lang import override
from aiida.engine import ExitCode, ProcessSpec, while_
from aiida.engine.processes.workchains.restart import BaseRestartWorkChain, ProcessHandlerReport, process_handler
from aiida.plugins import CalculationFactory

# pylint: disable=too-many-branches, too-many-statements
from packaging import version

from aiida_vasp.assistant.parameters import ParametersMassage
from aiida_vasp.calcs.neb import VaspNEBCalculation
from aiida_vasp.data.potcar import PotcarData
from aiida_vasp.parsers.content_parsers.potcar import MultiPotcarIo
from aiida_vasp.utils.workchains import compose_exit_code

# Additional tags for VTST calculations - these are not the tags used by standard VASP
VTST_ADDITIONAL_TAGS = {
    'iopt': 'TAG for VTST',
    'maxmove': 'Maximum ionic movement',
    'ilbfgsmem': 'Number of steps saved when building the inverse Hessian matrix',
    'lglobal': 'Optimize the NEB globally instead of image-by-image',
    'lautoscale': 'Automatically determines INVCURV',
    'invcurv': 'Initial inverse curvature, used to construct the inverse Hessian matrix',
    'llineopt': 'Use a force based line minimizer for translation',
    'fdstep': 'Finite difference step size for line optimizer',
    'timestep': 'Dynamical time step',
    'sdalpha': 'Ratio between force and step size',
    'ftimemax': 'Maximum dynamical time step allowed',
    'ftimedec': 'Factor to decrease dt',
    'ftimeinc': 'Factor to increase dt',
    'falpha': 'Parameter that controls velocity damping',
    'fnmin': 'Minium number of iterations before adjust alpha and dt',
    'lclimb': 'Use climbing image mode',
    'ichain': 'Indicates which method to run. NEB (ICHAIN=0) is the default',
    'ltangentold': 'Flag to turn on the old central difference tangent',
    'ldneb': 'Flag to turn on modified double nudging',
    'lnebcell': 'Flag to turn on SS-NEB. Used with ISIF=3 and IOPT=3.',
    'jacobian': 'Controls weight of lattice to atomic motion. Ω is volume and N is the number of atoms.',
}


class VaspNEBWorkChain(BaseRestartWorkChain):
    """
    A NEB workchain

    Error handling enriched wrapper around VaspNEBCalculation.

    Deliberately conserves most of the interface (required inputs) of the VaspNEBCalculation class, but
    makes it possible for a user to interact with a workchain and not a calculation.

    In addition, implement restarts of calculation when the calculation is net full converged for error handling.

    """

    _verbose = False
    _process_class = CalculationFactory('vasp.neb')
    _norm_disp_threshold = 4.0

    @classmethod
    def define(cls, spec: ProcessSpec) -> None:
        super(VaspNEBWorkChain, cls).define(spec)
        spec.expose_inputs(
            cls._process_class,
            exclude=('potential', 'kpoints', 'dynamics', 'metadata'),
        )
        spec.input(
            'kpoints',
            valid_type=orm.KpointsData,
            required=False,
        )
        spec.input(
            'kpoints_spacing',
            valid_type=orm.Float,
            required=False,
        )
        spec.input(
            'potential_family',
            valid_type=orm.Str,
            required=True,
        )
        spec.input(
            'potential_mapping',
            valid_type=orm.Dict,
            required=True,
        )
        spec.input(
            'options',
            valid_type=orm.Dict,
            required=True,
        )
        spec.input(
            'max_iterations',
            valid_type=orm.Int,
            required=False,
            default=lambda: orm.Int(5),
            help="""
            The maximum number of iterations to perform.
            """,
        )
        spec.input(
            'clean_workdir',
            valid_type=orm.Bool,
            required=False,
            default=lambda: orm.Bool(False),
            help="""
            If True, clean the work dir upon the completion of a successful calculation.
            """,
        )
        spec.input(
            'verbose',
            valid_type=orm.Bool,
            required=False,
            default=lambda: orm.Bool(True),
            help="""
            If True, enable more detailed output during workchain execution.
            """,
        )
        spec.input(
            'dynamics.positions_dof',
            valid_type=orm.List,
            required=False,
            help="""
            Site dependent flag for selective dynamics when performing relaxation
            """,
        )
        spec.input(
            'ldau_mapping',
            valid_type=orm.Dict,
            required=False,
            help="Mappings, see the doc string of 'get_ldau_keys'",
        )
        spec.input(
            'kpoints_spacing',
            valid_type=orm.Float,
            required=False,
            help='Spacing for the kpoints in units A^-1 * 2pi (CASTEP style `kpoints_mp_spacing`)',
        )
        spec.input(
            'kpoints_spacing_vasp',
            valid_type=orm.Float,
            required=False,
            help='Spacing for the kpoints in units A^-1 (VASP style)',
        )
        spec.outline(
            cls.setup,
            while_(cls.should_run_process)(
                cls.run_process,
                cls.inspect_process,
            ),
            cls.results,
        )

        spec.expose_outputs(cls._process_class)
        spec.exit_code(
            0,
            'NO_ERROR',
            message='the sun is shining',
        )
        spec.exit_code(
            700,
            'ERROR_NO_POTENTIAL_FAMILY_NAME',
            message='the user did not supply a potential family name',
        )
        spec.exit_code(
            701,
            'ERROR_POTENTIAL_VALUE_ERROR',
            message='ValueError was returned from get_potcars_from_structure',
        )
        spec.exit_code(
            702,
            'ERROR_POTENTIAL_DO_NOT_EXIST',
            message='the potential does not exist',
        )
        spec.exit_code(
            703,
            'ERROR_IN_PARAMETER_MASSAGER',
            message='the exception: {exception} was thrown while massaging the parameters',
        )
        spec.exit_code(
            501,
            'SUB_NEB_CALCULATION_ERROR',
            message='Unrecoverable error in launched NEB calculations.',
        )

    def setup(self) -> None | ExitCode:
        super().setup()

        # Setup the initial inputs
        self.ctx.inputs = self.exposed_inputs(self._process_class)

        # Stage the neb images
        self.ctx.neb_images = self.inputs.neb_images

        # Handle and convert additional inputs and store them in self.ctx.inputs
        exit_code = self._setup_vasp_inputs()
        if exit_code is not None:
            return exit_code

        # Sanity checks
        self._check_neb_inputs()
        return None

    @process_handler(priority=500, exit_codes=[VaspNEBCalculation.exit_codes.ERROR_IONIC_NOT_CONVERGED])  # pylint: disable=no-member
    def handle_unconverged(self, node: orm.WorkChainNode) -> ProcessHandlerReport | None:
        """
        Handle the problem where the NEB optimization is not converged.

        Note that VASP could reach NSW before the actual convergence.
        Hence this check is necessary even for finished runs.
        """
        if 'misc' not in node.outputs:
            self.report('Cannot found the `misc` output containing the NEB run data')
            return None
        misc_dict = node.outputs.misc.get_dict()

        neb_data = misc_dict.get('neb_data')
        if neb_data is None:
            self.report('Cannot found the `neb_data` dictionary containing the NEB run data')
            return None

        converged = [tmp.get('neb_converged', False) for tmp in neb_data.values()]
        if not all(converged):
            self.report('At least one image is not converged in the run. Restart required.')

            # Attach images
            out = self._attach_output_structure(node)
            if out is not None:
                return out

            self.report(f'Successfully handled unconverged calculation {node}.')
            return ProcessHandlerReport()
        self.report(f'Cannot handle ionically unconverged calculation {node}.')
        return None

    @process_handler(priority=900, exit_codes=[VaspNEBCalculation.exit_codes.ERROR_DID_NOT_FINISH])  # pylint: disable=no-member
    def handle_unfinished(self, node: orm.WorkChainNode) -> ProcessHandlerReport | None:
        """
        Handle the case where the calculations is not fully finished.
        This checks the existing of the run_stats field in the parsed per-image misc output
        """

        finished = []
        # Since 1.6.3 the nested namespaces are handled properly.
        if version.parse(aiida_version) >= version.parse('1.6.3'):
            if 'misc' not in node.outputs:
                self.report('Cannot found the `misc` output containing the parsed per-image data')
                return None

            misc_dict = node.outputs.misc.get_dict()
            if 'run_status' in misc_dict:
                finished = {key: value.get('finished', False) for key, value in misc_dict['run_status'].items()}

        if not all(finished.values()):
            self.report('At least one image did not reach the end of VASP execution - calculation not finished!')

            out = self._attach_output_structure(node)
            if out is not None:
                return out

            # No further process handling is needed
            self.report(f'Successfully handled unfinished calculation {node}.')
            return ProcessHandlerReport(do_break=True)
        self.report(f'Cannot handle unfinished calculation {node}.')
        return None

    def _attach_output_structure(self, node: orm.WorkChainNode) -> ProcessHandlerReport | None:
        """
        Attached the output structure of a children node as the inputs for the
        next workchain launch.
        """
        output_images = AttributeDict()  # A dictionary holding the structures with keys like 'image_xx'
        output_images = node.outputs['structure']

        nout = len(output_images)
        nexists = len(self.inputs.neb_images)
        if nout != nexists:
            self.report(f'Number of parsed images: {nout} does not equal to the images need to restart: {nexists}.')
            return ProcessHandlerReport(do_break=True, exit_code=self.exit_codes.SUB_NEB_CALCULATION_ERROR)  # pylint: disable=no-member
        self.report(f'Attached output structures from the previous calculation {node} as new inputs.')
        self.ctx.inputs.neb_images = output_images
        return None

    def _check_neb_inputs(self) -> None:
        """
        Perform some simple checks for the NEB inputs

        This method is called once by ``self.setup``
        """

        incar = self.ctx.inputs.parameters

        images = incar.get('images')

        if not images:
            raise InputValidationError('IMAGES parameters is not set in the INCAR inputs')

        nimages = len(self.ctx.inputs.neb_images)

        if nimages != images:
            raise InputValidationError('Mismatch between IMAGES and actual number supplied input structures.')

        # Check for NEB tags
        iopt = incar.get('iopt', 0)
        ibrion = incar.get('ibrion')
        potim = incar.get('potim')

        # Check the sanity of parameters
        if ibrion != 3:
            self.report('WARNING: IBRION should be set to 3 for VTST runs, proceed with caution.')
        elif potim != 0:
            self.report(
                'WARNING: Using VTST optimisors with IBRION=3, but POTIM is not set to zero, proceed with caution.'
            )
        if iopt == 0:
            self.report('WARNING: IOPT not set.')

        if ibrion == 2:
            raise InputValidationError('IBRION=2 should not be used for NEB optimization!!')

        # Check the displacement of atoms between the frames
        # the hope is that this may detect simple errors such as atoms going across the PBC or
        # the order of atoms are changed between different frames

        tmp = list(self.ctx.inputs.neb_images.items())
        tmp.sort(key=lambda x: x[0])
        frames = [x[1].get_ase() for x in tmp]
        frames = [self.ctx.inputs.initial_structure.get_ase()] + frames + [self.ctx.inputs.final_structure.get_ase()]

        last_frame = frames[0]
        # Function for computing the distance using the scaled positions
        rel_dist = np.vectorize(lambda x: x if x < 0.5 else 1.0 - x)
        for iframe, frame in enumerate(frames[1:]):
            # Relative displacements
            disp = abs(frame.get_scaled_positions() - last_frame.get_scaled_positions()) % 1.0
            # Apply convention
            disp = rel_dist(disp)
            # Convert back to absolute displacement
            disp = disp @ frame.cell
            norm_disp = np.linalg.norm(disp, axis=1)
            sort_idx = np.argsort(norm_disp)
            if norm_disp[sort_idx[-1]] > self._norm_disp_threshold:
                raise InputValidationError(
                    'Large displacement detected for atom {} at frame {} - please check the inputs images'.format(
                        sort_idx[-1], iframe + 1
                    )
                )
            last_frame = frame

    def _setup_vasp_inputs(self) -> None:
        """
        Setup the inputs for VASP calculation

        This method is called once by ``self.setup``
        TODO: merge with vasp.v2.vasp
        """

        # Set the kpoints (kpoints)
        if 'kpoints' in self.inputs:
            self.ctx.inputs.kpoints = self.inputs.kpoints
        elif 'kpoints_spacing' in self.inputs:
            kpoints = orm.KpointsData()
            kpoints.set_cell_from_structure(self.ctx.inputs.initial_structure)
            kpoints.set_kpoints_mesh_from_density(self.inputs.kpoints_spacing.value * np.pi * 2)
            self.ctx.inputs.kpoints = kpoints
        elif 'kpoints_spacing_vasp' in self.inputs:
            kpoints = orm.KpointsData()
            kpoints.set_cell_from_structure(self.ctx.inputs.initial_structure)
            kpoints.set_kpoints_mesh_from_density(self.inputs.kpoints_spacing.value)
            self.ctx.inputs.kpoints = kpoints
        else:
            raise InputValidationError("Must supply either 'kpoints' or 'kpoints_spacing' or 'kpoints_spacing_vasp")

        # Set settings

        unsupported_parameters = dict(VTST_ADDITIONAL_TAGS)
        if 'settings' in self.inputs:
            self.ctx.inputs.settings = self.inputs.settings
            # Also check if the user supplied additional tags that is not in the supported file.
            try:
                unsupported_parameters = self.ctx.inputs.settings.unsupported_parameters
            except AttributeError:
                pass

        # Perform inputs massage to accommodate generalization in higher lying workchains
        # and set parameters.
        try:
            parameters_massager = ParametersMassage(self.inputs.parameters, unsupported_parameters)
        except Exception as exception:  # pylint: disable=broad-except
            return self.exit_codes.ERROR_IN_PARAMETER_MASSAGER.format(exception=exception)  # pylint: disable=no-member
        try:
            # Only set if they exists
            # Set any INCAR tags
            self.ctx.inputs.parameters = parameters_massager.parameters.incar
            # Set any dynamics input (currently only for selective dynamics, e.g. custom write to POSCAR)
            self.ctx.inputs.dynamics = parameters_massager.parameters.dynamics
            # Here we could set additional override flags, but those are not relevant for this VASP plugin
        except AttributeError:
            pass

        # Setup LDAU keys
        if 'ldau_mapping' in self.inputs:
            ldau_settings = self.inputs.ldau_mapping.get_dict()
            ldau_keys = get_ldau_keys(self.ctx.inputs.initial_structure, **ldau_settings)
            # Directly update the raw inputs passed to VaspCalculation
            self.ctx.inputs.parameters.update(ldau_keys)

        # Set settings
        if 'settings' in self.inputs:
            self.ctx.inputs.settings = self.inputs.settings

        # Set options
        # Options is very special, not storable and should be
        # wrapped in the metadata dictionary, which is also not storable
        # and should contain an entry for options
        if 'options' in self.inputs:
            options = {}
            options.update(self.inputs.options)
            self.ctx.inputs.metadata = {}
            self.ctx.inputs.metadata['options'] = options
            # Override the parser name if it is supplied by the user.
            parser_name = self.ctx.inputs.metadata['options'].get('parser_name')
            if parser_name:
                self.ctx.inputs.metadata['options']['parser_name'] = parser_name
            # Also make sure we specify the entry point for the
            # Set MPI to True, unless the user specifies otherwise
            withmpi = self.ctx.inputs.metadata['options'].get('withmpi', True)
            self.ctx.inputs.metadata['options']['withmpi'] = withmpi
        else:
            raise InputValidationError('`options` not supplied')

        # Utilise default input/output selections
        self.ctx.inputs.metadata['options']['input_filename'] = 'INCAR'

        # Set the CalcJobNode to have the same label as the WorkChain
        self.ctx.inputs.metadata['label'] = self.inputs.metadata.get('label', '')
        self.report(self.ctx.inputs.metadata)

        # Verify and set potentials (potcar)
        if not self.inputs.potential_family.value:
            self.report(  # pylint: disable=not-callable
                'An empty string for the potential family name was detected.'
            )
            return self.exit_codes.ERROR_NO_POTENTIAL_FAMILY_NAME  # pylint: disable=no-member
        try:
            self.ctx.inputs.potential = PotcarData.get_potcars_from_structure(
                structure=self.inputs.initial_structure,
                family_name=self.inputs.potential_family.value,
                mapping=self.inputs.potential_mapping.get_dict(),
            )
        except ValueError as err:
            return compose_exit_code(self.exit_codes.ERROR_POTENTIAL_VALUE_ERROR.status, str(err))  # pylint: disable=no-member
        except NotExistent as err:
            return compose_exit_code(self.exit_codes.ERROR_POTENTIAL_DO_NOT_EXIST.status, str(err))  # pylint: disable=no-member

        self.ctx.verbose = bool(self.inputs.get('verbose', self._verbose))

        return None

    @override
    def results(self) -> ExitCode | None:
        """Attach the outputs specified in the output specification from the last completed process."""
        node = self.ctx.children[self.ctx.iteration - 1]

        # We check the `is_finished` attribute of the work chain and not the successfulness of the last process
        # because the error handlers in the last iteration can have qualified a "failed" process as satisfactory
        # for the outcome of the work chain and so have marked it as `is_finished=True`.
        max_iterations = self.inputs.max_iterations.value  # type: ignore[union-attr]
        if not self.ctx.is_finished and self.ctx.iteration >= max_iterations:
            self.report(
                f'reached the maximum number of iterations {max_iterations}: '
                f'last ran {self.ctx.process_name}<{node.pk}>'
            )
            return self.exit_codes.ERROR_MAXIMUM_ITERATIONS_EXCEEDED  # pylint: disable=no-member

        self.report(f'work chain completed after {self.ctx.iteration} iterations')

        # Simply attach the output of the last children
        self.out_many({key: node.outputs[key] for key in node.outputs})
        return None


# The code below should be moved for utility module, but I keep them here for now

FELEMS = [
    'La',
    'Ce',
    'Pr',
    'Nd',
    'Pm',
    'Sm',
    'Eu',
    'Gd',
    'Tb',
    'Dy',
    'Ho',
    'Er',
    'Tm',
    'Yb',
    'Lu',
    'Ac',
    'Th',
    'Pa',
    'U',
    'Np',
    'Pu',
    'Am',
    'Cm',
    'Bk',
    'Cf',
    'Es',
    'Fm',
    'Md',
    'No',
    'Lr',
]


def get_ldau_keys(
    structure: orm.StructureData,
    mapping: dict[str, list[float]],
    utype: int = 2,
    jmapping: dict[str, float] | None = None,
    felec: bool = False,
) -> dict[str, Any]:
    """
    Setup LDAU mapping.

    In VASP, the U for each species has to be defined in the order that they appear in POSCAR.
    This is a helper to make sure the values of U are associated to each species.

    :param structure: The structure, either StructureData or ase.Atoms is fine.
    :param mapping: A dictionary in the format of  ``{"Mn": [d, 4]...}`` for U.
    :param utype: The type of LDA+U, default to 2, which is the one with only one parameter.
    :param jmapping: A dictionary in the format of  ``{"Mn": [d, 4]...}`` but for J.
    :param felec: Whether we are dealing with f electrons; will increase lmaxmix if we are.

    :returns: A dictionary to be used to update the raw input parameters for VASP.
    """
    if isinstance(structure, orm.StructureData):
        species = MultiPotcarIo.potentials_order(structure)
    else:
        # For ASE atoms, we keep the order of species occurrence no sorting is done
        species = []
        for symbol in structure.get_chemical_symbols():
            if symbol not in species:
                species.append(symbol)

    lsymbols = {'d': 2, 'f': 3, 'p': 1}
    if jmapping is None:
        jmapping = {}
    # Setup U array
    ldauu = []
    ldauj = []
    ldaul = []
    count = 0
    for specie in species:
        if specie in mapping:
            uvalue = mapping[specie][1]
            j = jmapping.get(specie, 0.0)
            ldaul.append(lsymbols[mapping[specie][0]])
            ldauu.append(mapping[specie][1])

            j = jmapping.get(specie, 0.0)
            ldauj.append(j)

            if specie in FELEMS:
                felec = True
            # Count the number of valid mappings
            if uvalue != 0.0 or j != 0.0:
                count += 1

        else:
            ldauu.append(0.0)
            ldauj.append(0.0)
            ldaul.append(-1)

    if count > 0:
        # Only enable U is there is any non-zero value
        output = {
            'ldauu': ldauu,
            'ldauj': ldauj,
            'ldautype': utype,
            'lmaxmix': 6 if felec else 4,
            'ldaul': ldaul,
            'ldau': True,
        }
    else:
        output = {}
    return output



"""
Test for input set specifications
"""

import numpy as np
import pytest
from aiida import orm
from ase.build import bulk

from aiida_vasp.workchains.v2.inputset.base import InputSet
from aiida_vasp.workchains.v2.inputset.pmgset import PymatgenInputSet
from aiida_vasp.workchains.v2.inputset.vaspsets import VASPInputSet

try:
    # Import SETTINGS
    from pymatgen.core import SETTINGS

    # Check equivalence with pymatgen
    from pymatgen.io.vasp.sets import MPRelaxSet
except ImportError:
    MPRelaxSet = None

# pylint:disable=redefined-outer-name,unused-argument


@pytest.fixture
def fe_atoms(aiida_profile):
    """Get a Fe atoms"""
    return orm.StructureData(ase=bulk('Fe', 'fcc', 5.0))


@pytest.fixture
def feo_atoms(aiida_profile):
    """Get a FeO atoms in rocksalt structure"""
    return orm.StructureData(ase=bulk('FeO', 'rocksalt', 5.0))


@pytest.fixture
def mgo_atoms(aiida_profile):
    """Get a MgO atoms"""
    return orm.StructureData(ase=bulk('MgO', 'rocksalt', 5.0))


def test_base(fe_atoms):
    """Base test case"""
    iset = InputSet('MITRelaxSet', overrides={'ediff': 1.0, 'nsw': None})

    out = iset.get_input_dict(fe_atoms)
    assert out['ediff'] == 1.0
    assert out['ibrion'] == 2
    assert 'nsw' not in out


def test_vasp(fe_atoms):
    """Test VASP inputsets"""
    iset = VASPInputSet('MITRelaxSet', overrides={'ediff': 1.0, 'nsw': None, 'ldautype': 3})

    out = iset.get_input_dict(fe_atoms)
    assert out['ediff'] == 1.0
    assert out['ibrion'] == 2
    assert out['magmom'] == [5]
    assert out['ldauu'] == [4.0]
    assert out['ldauj'] == [0.0]
    assert out['ldaul'] == [2]
    assert out['ldautype'] == 3
    assert out['ldau'] is True
    assert 'nsw' not in out


def test_kpoints(aiida_profile, fe_atoms):
    """Test generating kpoints"""
    inset = VASPInputSet('MITRelaxSet')
    kpoints = inset.get_kpoints(fe_atoms, 0.05)
    assert kpoints.get_kpoints_mesh()[0][0] == 7


@pytest.mark.skipif(MPRelaxSet is None, reason='pymatgen is not installed')
def test_pmg_kpoints(aiida_profile, fe_atoms):
    """Test generating kpoints"""
    inset = PymatgenInputSet('MPRelaxSet')
    kpoints = inset.get_kpoints(fe_atoms)
    assert kpoints.get_kpoints_mesh()[0][0] == 7
    assert kpoints.get_kpoints_mesh()[1][0] == 0

    inset = PymatgenInputSet('MPRelaxSet')
    scaled_fe_atoms = fe_atoms.get_ase()
    scaled_fe_atoms.set_cell(scaled_fe_atoms.get_cell() * 1.2)
    # Breaks the symmetry so it is not face centred
    scaled_fe_atoms.cell += np.random.rand(3, 3) * 0.1
    scaled_fe_atoms = orm.StructureData(ase=scaled_fe_atoms)
    kpoints = inset.get_kpoints(scaled_fe_atoms)
    assert kpoints.get_kpoints_mesh()[0][0] == 6
    assert kpoints.get_kpoints_mesh()[1][0] == -0.5

    inset = PymatgenInputSet('MP24RelaxSet')
    kpoints = inset.get_kpoints(fe_atoms)
    assert kpoints is None
    kspacing = inset.get_kpoints_spacing(fe_atoms)
    assert kspacing == 0.22 / np.pi / 2


@pytest.mark.skipif(MPRelaxSet is None, reason='pymatgen is not installed')
def test_pmg(fe_atoms, feo_atoms):
    """Test pymatgen inputsets"""
    vpmgset = PymatgenInputSet('MPRelaxSet', overrides={'ediff': 1.0, 'nsw': None, 'ldautype': 3})
    out = vpmgset.get_input_dict(fe_atoms)
    assert 'nsw' not in out
    assert out['ediff'] == 1.0
    assert out['ldautype'] == 3
    assert 'ldau' not in out
    assert out['ismear'] == -5

    vpmgset = PymatgenInputSet('MPRelaxSet', overrides={'ediff': 1.0, 'nsw': None})
    out = vpmgset.get_input_dict(feo_atoms)
    assert 'ldau' in out

    psp_dir = SETTINGS.get('PMG_VASP_PSP_DIR')
    for structure in feo_atoms, fe_atoms:
        pmgset = MPRelaxSet(structure.get_pymatgen())
        vpmgset = PymatgenInputSet('MPRelaxSet')
        ref_dict = dict(pmgset.incar)
        ref_dict.pop('icharg', None)
        ref_dict.pop('istart', None)
        ref_dict.pop('kspacing', None)
        out = vpmgset.get_input_dict(structure)
        assert ref_dict == {key.upper(): value for key, value in out.items()}
        assert pmgset.kpoints.kpts[0] == tuple(vpmgset.get_kpoints(structure).get_kpoints_mesh()[0])
        if psp_dir:
            assert {p.element: p.symbol for p in pmgset.potcar} == vpmgset.get_pp_mapping(structure)



"""
Collection of process functions for AiiDA, used for structure transformation
"""

import re
from typing import Any, List, Tuple

import numpy as np
from aiida import orm
from aiida.engine import calcfunction
from aiida.orm import StructureData
from aiida.tools.data.array.kpoints import get_kpoints_path
from aiida.tools.data.structure import spglib_tuple_to_structure, structure_to_spglib_tuple
from ase import Atoms
from ase.build import niggli_reduce as niggli_reduce_
from ase.build import sort
from ase.build.supercells import make_supercell as ase_supercell
from ase.neb import NEB
from spglib import niggli_reduce as niggli_reduce_spg
from spglib import refine_cell


@calcfunction
def magnetic_structure_decorate(structure: orm.StructureData, magmom: orm.List) -> dict[str, Any]:
    """
    Create Quantum Espresso style decorated structure with
    given magnetic moments.
    """

    magmom = magmom.get_list()
    assert len(magmom) == len(structure.sites), (
        f'Mismatch between the magmom ({len(magmom)}) and the nubmer of sites ({len(structure.sites)}).'
    )
    old_species = [structure.get_kind(site.kind_name).symbol for site in structure.sites]
    new_species, magmom_mapping = create_additional_species(old_species, magmom)
    new_structure = StructureData()
    new_structure.set_cell(structure.cell)
    new_structure.set_pbc(structure.pbc)
    for site, name in zip(structure.sites, new_species):
        this_symbol = structure.get_kind(site.kind_name).symbol
        new_structure.append_atom(position=site.position, symbols=this_symbol, name=name)

    # Keep the label
    new_structure.label = structure.label
    return {'structure': new_structure, 'mapping': orm.Dict(dict=magmom_mapping)}


@calcfunction
def magnetic_structure_dedecorate(structure: orm.StructureData, mapping: orm.Dict) -> dict[str, Any]:
    """
    Remove decorations of a structure with multiple names for the same specie
    given that the decoration was previously created to give different species
    name for different initialisation of magnetic moments.
    """

    mapping = mapping.get_dict()
    # Get a list of decroated names
    old_species = [structure.get_kind(site.kind_name).name for site in structure.sites]
    new_species, magmom = convert_to_plain_list(old_species, mapping)

    new_structure = StructureData()
    new_structure.set_cell(structure.cell)
    new_structure.set_pbc(structure.pbc)

    for site, name in zip(structure.sites, new_species):
        this_symbol = structure.get_kind(site.kind_name).symbol
        new_structure.append_atom(position=site.position, symbols=this_symbol, name=name)
    new_structure.label = structure.label
    return {'structure': new_structure, 'magmom': orm.List(list=magmom)}


@calcfunction
def rattle(structure: orm.StructureData, amp: orm.Float) -> orm.StructureData:
    """
    Rattle the structure by a certain amplitude
    """
    native_keys = ['cell', 'pbc1', 'pbc2', 'pbc3', 'kinds', 'sites', 'mp_id']
    # Keep the foreign keys as it is
    foreign_attrs = {key: value for key, value in structure.attributes.items() if key not in native_keys}
    atoms = structure.get_ase()
    atoms.rattle(amp.value)
    # Clean any tags etc
    atoms.set_tags(None)
    atoms.set_masses(None)
    # Convert it back
    out = StructureData(ase=atoms)
    out.base.attributes.set_many(foreign_attrs)
    out.label = structure.label + ' RATTLED'
    return out


@calcfunction
def get_primitive(structure: orm.StructureData) -> orm.StructureData:
    """Create primitive structure use pymatgen interface"""
    pstruct = structure.get_pymatgen()
    ps = pstruct.get_primitive_structure()
    out = StructureData(pymatgen=ps)
    out.label = structure.label + ' PRIMITIVE'
    return out


@calcfunction
def get_standard_primitive(structure: orm.StructureData, **kwargs: Any) -> orm.StructureData:
    """Create the standard primitive structure via seekpath"""

    parameters = kwargs.get('parameters', {'symprec': 1e-5})
    if isinstance(parameters, orm.Dict):
        parameters = parameters.get_dict()

    out = get_kpoints_path(structure, **parameters)['primitive_structure']
    out.label = structure.label + ' PRIMITIVE'
    return out


@calcfunction
def spglib_refine_cell(structure: orm.StructureData, symprec: Any) -> orm.StructureData:
    """Create the standard primitive structure via seekpath"""

    structure_tuple, kind_info, kinds = structure_to_spglib_tuple(structure)

    lattice, positions, types = refine_cell(structure_tuple, symprec.value)

    refined = spglib_tuple_to_structure((lattice, positions, types), kind_info, kinds)

    return refined


@calcfunction
def get_standard_conventional(structure: orm.StructureData) -> orm.StructureData:
    """Create the standard primitive structure via seekpath"""
    out = get_kpoints_path(structure)['conv_structure']
    out.label = structure.label + ' PRIMITIVE'
    return out


@calcfunction
def get_refined_structure(structure: orm.StructureData, symprec: orm.Float, angle_tolerance: Any) -> orm.StructureData:
    """Create refined structure use pymatgen's interface"""
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer  # noqa: PLC0415

    pstruct = structure.get_pymatgen()
    ana = SpacegroupAnalyzer(pstruct, symprec=symprec.value, angle_tolerance=angle_tolerance.value)
    ps = ana.get_refined_structure()
    out = StructureData(pymatgen=ps)
    out.label = structure.label + ' REFINED'
    return out


@calcfunction
def get_conventional_standard_structure(structure: orm.StructureData, symprec: orm.Float, angle_tolerance: Any) -> Any:
    """Create conventional standard structure use pymatgen's interface"""
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer  # noqa: PLC0415

    pstruct = structure.get_pymatgen()
    ana = SpacegroupAnalyzer(pstruct, symprec=symprec.value, angle_tolerance=angle_tolerance.value)
    ps = ana.get_conventional_standard_structure()
    out = StructureData(pymatgen=ps)
    out.label = structure.label + ' CONVENTIONAL STANDARD'
    return out


@calcfunction
def make_supercell(structure: Any, supercell: Any, **kwargs: Any) -> dict[str, Any]:
    """Make supercell structure, keep the tags in order"""

    if 'tags' in kwargs:
        tags = kwargs['tags']
    else:
        tags = None

    atoms = structure.get_ase()
    atoms.set_tags(tags)

    slist = supercell.get_list()
    if isinstance(slist[0], int):
        satoms = atoms.repeat(slist)
    else:
        satoms = ase_supercell(atoms, np.array(slist))
    if 'no_sort' not in kwargs:
        satoms = sort(satoms)

    if tags:
        stags = satoms.get_tags().tolist()
    satoms.set_tags(None)

    out = StructureData(ase=satoms)
    out.label = structure.label + ' SUPER {} {} {}'.format(*slist)

    if tags:
        return {'structure': out, 'tags': orm.List(list=stags)}
    else:
        return {'structure': out}


@calcfunction
def niggli_reduce(structure: orm.StructureData) -> orm.StructureData:
    """Peroform niggli reduction using ase as the backend - this will rotate the structure into the standard setting"""
    atoms = structure.get_ase()
    niggli_reduce_(atoms)
    new_structure = StructureData(ase=atoms)
    new_structure.label = structure.label + ' NIGGLI'
    return new_structure


@calcfunction
def niggli_reduce_spglib(structure: orm.StructureData) -> orm.StructureData:
    """Peroform niggli reduction using spglib as backend - this does not rotate the structure"""
    atoms = structure.get_ase()
    reduced_cell = niggli_reduce_spg(atoms.cell)
    atoms.set_cell(reduced_cell)
    atoms.wrap()
    new_structure = StructureData(ase=atoms)
    new_structure.label = structure.label + ' NIGGLI'
    return new_structure


@calcfunction
def neb_interpolate(
    init_structure: orm.StructureData, final_strucrture: orm.StructureData, nimages: orm.Int
) -> dict[str, orm.StructureData]:
    """
    Interpolate NEB frames using the starting and the final structures

    Get around the PBC wrapping problem by calculating the MIC displacements
    from the initial to the final structure

    The initial structure is not changed, while the final structure is
    modified to be consistent with the initial structure in terms of absolute
    displacements, i.e. the final structure is *unwrapped*.
    """
    ainit = init_structure.get_ase()
    afinal = final_strucrture.get_ase()
    disps = []

    # Find distances
    acombined = ainit.copy()
    acombined.extend(afinal)
    # Get piece-wise MIC distances
    for i in range(len(ainit)):
        dist = acombined.get_distance(i, i + len(ainit), vector=True, mic=True)
        disps.append(dist.tolist())
    disps = np.asarray(disps)
    afinal = ainit.copy()

    # Displace the atoms according to MIC distances
    afinal.positions += disps
    neb = NEB([ainit.copy() for i in range(int(nimages) + 1)] + [afinal.copy()])
    neb.interpolate()
    out_init = StructureData(ase=neb.images[0])
    out_init.label = init_structure.label + ' INIT'
    out_final = StructureData(ase=neb.images[-1])
    out_final.label = init_structure.label + ' FINAL'

    outputs = {'image_init': out_init}
    for i, out in enumerate(neb.images[1:-1]):
        outputs[f'image_{i + 1:02d}'] = StructureData(ase=out)
        outputs[f'image_{i + 1:02d}'].label = init_structure.label + f' FRAME {i + 1:02d}'
    outputs['image_final'] = out_final
    return outputs


@calcfunction
def fix_atom_order(reference: orm.StructureData, to_fix: orm.StructureData) -> orm.StructureData:
    """
    Fix atom order by finding NN distances bet ween two frames. This resolves
    the issue where two closely matching structures having diffferent atomic orders.
    Note that the two frames must be close enough for this to work
    """

    aref = reference.get_ase()
    afix = to_fix.get_ase()

    # Index of the reference atom in the second structure
    new_indices = np.zeros(len(aref), dtype=int)

    # Find distances
    acombined = aref.copy()
    acombined.extend(afix)
    # Get piece-wise MIC distances
    for i in range(len(aref)):
        dists = []
        for j in range(len(aref)):
            dist = acombined.get_distance(i, j + len(aref), mic=True)
            dists.append(dist)
        min_idx = np.argmin(dists)
        min_dist = min(dists)
        if min_dist > 0.5:
            print(f'Large displacement found - moving atom {j} to {i} - please check if this is correct!')
        new_indices[i] = min_idx

    afixed = afix[new_indices]
    fixed_structure = StructureData(ase=afixed)
    fixed_structure.label = to_fix.label + ' UPDATED ORDER'
    return fixed_structure


def match_atomic_order_(atoms: Atoms, atoms_ref: Atoms) -> Tuple[Atoms, List[int]]:
    """
    Reorder the atoms to that of the reference.

    Only works for identical or nearly identical structures that are ordered differently.
    Returns a new `Atoms` object with order similar to that of `atoms_ref` as well as the sorting indices.
    """

    # Find distances
    acombined = atoms_ref.copy()
    acombined.extend(atoms)
    new_index = []
    # Get piece-wise MIC distances
    jidx = list(range(len(atoms), len(atoms) * 2))
    for i in range(len(atoms)):
        dists = acombined.get_distances(i, jidx, mic=True)
        # Find the index of the atom with the smallest distance
        min_idx = np.where(dists == dists.min())[0][0]
        new_index.append(min_idx)
    assert len(set(new_index)) == len(atoms), 'The detected mapping is not unique!'
    return atoms[new_index], new_index


def create_additional_species(species: list[str], magmoms: list[float]) -> tuple[list[str], dict[str, float]]:
    """
    Create additional species depending on magnetic moments.
    For example, create Fe1 and Fe2 if there are Fe with different
    magnetisations.

    :return: a tuples of (newspecies, magmom_mapping)
    """

    unique_species: set[str] = set(species)
    new_species: list[str] = []
    current_species_mapping: dict[str, dict[str, float]] = {sym: {} for sym in unique_species}
    for symbol, magmom in zip(species, magmoms):
        current_symbol: str = symbol
        # Mappings for this original symbol
        mapping: dict[str, float] = current_species_mapping[symbol]
        # First check if this magmom has been treated
        not_seen = True
        for sym_, mag_ in mapping.items():
            if mag_ == magmom:
                current_symbol = sym_
                not_seen = False
        # This symbol has not been seen yet
        if not_seen:
            if current_symbol in mapping:
                # The other species having the same symbol has been assigned
                counter = len(mapping) + 1
                current_symbol = f'{symbol}{counter}'
            mapping[current_symbol] = magmom
        new_species.append(current_symbol)

    # Rename symbols that has more than one species, so A becomes A1
    for symbol, mapping in current_species_mapping.items():
        if len(mapping) > 1:
            mapping[f'{symbol}1'] = mapping[symbol]
            mapping.pop(symbol)
            # Refresh the new_species list
            new_species = [f'{sym}1' if sym == symbol else sym for sym in new_species]

    all_mapping: dict[str, float] = {}
    for value in current_species_mapping.values():
        all_mapping.update(value)

    return new_species, all_mapping


def convert_to_plain_list(species: list[str], magmom_mapping: dict[str, float]) -> tuple[list[str], list[float]]:
    """
    Covert from a decorated species list to a plain list of symbols
    and magnetic moments.

    :return: A tuple of (symbols, magmoms)
    """
    magmoms = []
    symbols = []
    for symbol in species:
        magmoms.append(magmom_mapping[symbol])
        match = re.match(r'(\w+)\d+', symbol)
        if match:
            symbols.append(match.group(1))
        else:
            symbols.append(symbol)
    return symbols, magmoms



"""
Relaxation workchain for VASP

Unfortunately, VASP does not check the convergence criteria properly:

- it only check *either* force or energy convergence between the last two iterations
- relaxation is performed with *constant basis set*, so a final singlepoint calculation is necessary if cell is to be
  relaxed
- no check for the convergence of the atomic positions
- no check about the convergence of the cell volume

Hence we have to control it externally to make sure the structures are properly relaxed,
and perform additional singlepoint calculation where necessary.

In addition, using this workchain, there is no need to set the IBRION, NSW, ISIF and ALGO tags explicitly,
the action can be controlled using the `relax_setting` input port. This will be merged with the `parameters` and
passed downstream to the `VaspBaseWorkChain` which will work out the correct combinations for the three.

CHANGELOG

0.2.1 - added `hybrid_calc_bootstrap` options
0.3.0 - make singpoint calculation reuse the `restart_folder`
0.3.1 - called processes now have link labels. Added customisable `settings` and `options` for the final relaxation.


"""

# pylint: disable=attribute-defined-outside-init

from __future__ import annotations

import numpy as np
from aiida import orm
from aiida.common.exceptions import InputValidationError
from aiida.common.extendeddicts import AttributeDict
from aiida.common.utils import classproperty
from aiida.engine import ExitCode, ProcessSpec, ToContext, WorkChain, append_, if_, while_
from aiida.orm.nodes.data.base import to_aiida_type
from aiida.plugins import WorkflowFactory

from aiida_vasp.utils.extended_dicts import update_nested_dict, update_nested_dict_node
from aiida_vasp.utils.opthold import RelaxOptions
from aiida_vasp.utils.workchains import compose_exit_code

from .common import OVERRIDE_NAMESPACE, site_magnetization_to_magmom
from .mixins import WithBuilderUpdater

__version__ = '0.5.0'

# Change log
# 0.4.0 update such `vasp` namespace in `parameters` is renamed to `incar`
# 0.5.0 update the logic of convergence checking. Cell comparsion is always done using the input/output structures.


class VaspRelaxWorkChain(WorkChain, WithBuilderUpdater):
    """Structure relaxation workchain."""

    _verbose: bool = True
    _base_workchain_string: str = 'vasp.v2.vasp'
    _base_workchain = WorkflowFactory(_base_workchain_string)
    option_class = RelaxOptions

    @classmethod
    def define(cls, spec: ProcessSpec) -> None:
        super().define(spec)
        spec.expose_inputs(cls._base_workchain, 'vasp', exclude=('structure',))
        spec.input('structure', valid_type=(orm.StructureData, orm.CifData))
        spec.input(
            'static_calc_parameters',
            valid_type=orm.Dict,
            required=False,
            serializer=to_aiida_type,
            help="""
                   The parameters (INCAR) to be used in the final static calculation.
                   """,
        )
        spec.input(
            'static_calc_settings',
            valid_type=orm.Dict,
            required=False,
            serializer=to_aiida_type,
            help="""
                   The full settings Dict to be used in the final static calculation.
                   """,
        )
        spec.input(
            'static_calc_options',
            valid_type=orm.Dict,
            required=False,
            serializer=to_aiida_type,
            help="""
                   The full options Dict to be used in the final static calculation.
                   """,
        )
        spec.input(
            'relax_settings',
            valid_type=orm.Dict,
            validator=RelaxOptions.aiida_validate,
            serializer=RelaxOptions.aiida_serialize,
            help=RelaxOptions.aiida_description(),
        )
        spec.input(
            'verbose',
            required=False,
            help='Increased verbosity.',
            valid_type=orm.Bool,
            serializer=to_aiida_type,
        )
        spec.exit_code(0, 'NO_ERROR', message='the sun is shining')
        spec.exit_code(
            300,
            'ERROR_MISSING_REQUIRED_OUTPUT',
            message='the called workchain does not contain the necessary relaxed output structure',
        )
        spec.exit_code(420, 'ERROR_NO_CALLED_WORKCHAIN', message='no called workchain detected')
        spec.exit_code(
            500,
            'ERROR_UNKNOWN',
            message='unknown error detected in the relax workchain',
        )
        spec.exit_code(
            502,
            'ERROR_OVERRIDE_PARAMETERS',
            message='there was an error overriding the parameters',
        )
        spec.exit_code(
            600,
            'ERROR_RELAX_NOT_CONVERGED',
            message='Ionic relaxation was not converged after the maximum number of iterations has been spent',
        )
        spec.exit_code(
            601,
            'ERROR_FINAL_SCF_HAS_RESIDUAL_FORCE',
            message=(
                'The final singlepoint calculation has increased residual forces. This'
                ' may be caused by electronic solver converging to a different'
                ' solution. Care should be taken to investigate the results.'
            ),
        )
        spec.outline(
            cls.initialize,
            if_(cls.perform_relaxation)(
                while_(cls.run_next_relax)(
                    cls.run_relax,
                    cls.verify_last_relax,
                    cls.analyze_convergence,
                ),
                cls.store_relaxed,
            ),
            if_(cls.should_run_static_calculation)(
                cls.run_static_calculation,
                cls.verify_next_workchain,
            ),
            cls.results,
            cls.finalize,
        )  # yapf: disable

        spec.expose_outputs(cls._base_workchain)
        spec.output('relax.structure', valid_type=orm.StructureData, required=False)

    def initialize(self) -> None:
        """Initialize."""

        # Initialise the contexts
        self.ctx.exit_code = self.exit_codes.ERROR_UNKNOWN  # pylint: disable=no-member
        self.ctx.is_converged = False
        self.ctx.relax = False
        self.ctx.iteration = 0
        self.ctx.workchains = []
        self.ctx.inputs = AttributeDict()  # This may not be necessary anymore
        self.ctx.relax_settings = AttributeDict(
            self.inputs.relax_settings.get_dict()
        )  # relax_settings controls the logic of the workchain
        self.ctx.current_magmom = None
        self.ctx.is_double_relax = self.inputs.relax_settings.get('double_relax_mode', False)

        # Check potential issues in the the input parameters
        self._check_input_parameters()

        # Storage space for the inputs
        self.ctx.relax_input_additions = self._init_relax_input_additions()
        self.ctx.static_input_additions = AttributeDict()

        # Set the verbose flag
        if 'verbose' in self.inputs:
            self.ctx.verbose = self.inputs.verbose.value
        else:
            self.ctx.verbose = self._verbose  # Hard-coded default

        # Take the input structure as the "current" structure
        self.ctx.current_structure = self.inputs.structure

        # Make sure we parse the output structure when we want to perform
        # relaxations (override if contrary entry exists).
        if 'settings' in self.inputs.vasp:
            settings = self.inputs.vasp.settings
        else:
            settings = orm.Dict(dict={})

        if self.perform_relaxation():
            settings = update_nested_dict_node(
                settings,
                {'parser_settings': {'include_node': ['structure', 'trajectory'], 'include_quantity': ['energies']}},
                extend_list=True,
            )

        # Update the settings for the relaxation
        if settings.get_dict():
            self.ctx.relax_input_additions.settings = settings

        # Hybrid calculation boot-strapping
        if self.ctx.relax_settings.get('hybrid_calc_bootstrap'):
            self.ctx.hybrid_status = 'dft'
            if not self.ctx.relax_settings.get('reuse'):
                self.report('Enable `reuse` mode because hybrid calculation bootstrapping has been requested.')
                self.ctx.relax_settings['reuse'] = True
        else:
            self.ctx.hybrid_status = None

        # If we are reusing existing calculations, make sure the remote folders are not cleaned
        if self.ctx.relax_settings.get('reuse'):
            clean_tmp = self.inputs.vasp.get('clean_workdir')
            if clean_tmp and clean_tmp.value:
                self.report('Disable clean_workdir for downstream workflows since `reuse` is requested.')
                self.ctx.relax_input_additions['clean_workdir'] = orm.Bool(False)

        # Check the input parameters
        self._check_input_parameters()

    def _check_input_parameters(self) -> None:
        """Validate the input parameters and detect problems before running the workchain"""
        exposed = self.exposed_inputs(self._base_workchain, 'vasp')
        incar = exposed.parameters['incar']
        exp_key = ['ibrion', 'nsw', 'isif']
        for key in exp_key:
            if key in incar:
                self.report(
                    f'{key} explicitly set to {incar[key]} - this overrides the relax_settings input'
                    ' - proceed with caution.'
                )
        isif = incar.get('isif')
        if isif == 3 and not all(self.ctx.relax_settings.get(key) for key in ['positions', 'volume', 'shape']):
            raise InputValidationError(
                'ISIF = 3 is set explicity for INCAR, which is consistent with the mode of relaxation '
                'supplied to the workchain.'
            )

    def _init_relax_input_additions(self) -> AttributeDict:
        """
        Initialise the `relax_additions` field inside the context.
        It is a AttributeDict that contains the inputs that should be updated
        while performing the relaxation.
        """

        additions = AttributeDict()
        # Update the "relax" field inside the parameters - this is needed because some of the
        # settings will be translated into VASP parameters
        if self.perform_relaxation():
            parameters = update_nested_dict_node(self.inputs.vasp.parameters, {'relax': self.ctx.relax_settings})
            additions.parameters = parameters

        return additions

    def run_next_relax(self) -> bool:
        within_max_iterations = bool(self.ctx.iteration < self.ctx.relax_settings.convergence_max_iterations)
        return bool(within_max_iterations and not self.ctx.is_converged)

    def run_relax(self) -> ToContext:
        """Perform the relaxation"""
        # Iteration starts from 1, not 0
        self.ctx.iteration += 1

        inputs = self.exposed_inputs(self._base_workchain, 'vasp')
        inputs.structure = self.ctx.current_structure

        # Attach previous calculation's folder if requested
        if self.ctx.relax_settings.get('reuse', False):
            restart_folder = self.ctx.get('current_restart_folder')  # There might not be any yet
            if restart_folder:
                if self.ctx.get('verbose'):
                    self.report(f'Using previous remote folder <{restart_folder}> for restart')
                inputs.restart_folder = restart_folder

        # Update the input with whatever stored in the relax_input_additions attribute dict
        inputs.update(self.ctx.relax_input_additions)
        if 'label' not in inputs.metadata:
            inputs.metadata.label = self.inputs.metadata.get('label', '')

        # Check if we need to boot strap hybrid calculation
        if self.ctx.get('hybrid_status') == 'dft':
            # Turn off HF and turn off relaxation
            inputs.parameters = update_nested_dict_node(
                inputs.parameters,
                {
                    OVERRIDE_NAMESPACE: {
                        'lhfcalc': False,
                        'isym': 2,  # Standard DFT needs ISYM=2
                    },
                    'relax': {
                        # This turns off any relaxation
                        'positions': False,
                        'volume': False,
                        'shape': False,
                    },
                },
            )
            # Decrease the iteration number
            self.ctx.iteration -= 1
            # Update the wallclock seconds
            wallclock = self.ctx.relax_settings.get('hybrid_calc_bootstrap_wallclock')
            if wallclock:
                inputs.options = update_nested_dict_node(inputs.options, {'max_wallclock_seconds': wallclock})

        # Update the MAGMOM
        if self.ctx.current_magmom is not None:
            inputs.parameters = update_nested_dict_node(
                inputs.parameters,
                {OVERRIDE_NAMESPACE: {'magmom': self.ctx.current_magmom}},
            )

        # Label the calculation and links by iteration number
        inputs.metadata.label += f' ITER {self.ctx.iteration:02d}'
        inputs.metadata.call_link_label = f'relax_{self.ctx.iteration:02d}'

        # Keep the workdir if keep_sp_workdir is set to True but we are not having static calculation at all
        # This ensure that we have a valid remote_data output which is not cleaned by the called VaspWorkChain
        if self.inputs.relax_settings.get('keep_sp_workdir') and not self.should_run_static_calculation():
            inputs.keep_last_workdir = orm.Bool(True)

        running = self.submit(self._base_workchain, **inputs)
        self.report(f'launching {self._base_workchain.__name__}<{running.pk}> iterations #{self.ctx.iteration}')

        return ToContext(workchains=append_(running))

    def run_static_calculation(self) -> ToContext:
        """Perform the relaxation"""

        # For the final static run we do not need to parse the output structure
        if 'settings' in self.inputs.vasp:
            self.ctx.static_input_additions.settings = update_nested_dict_node(
                self.inputs.vasp.settings,
                {
                    'parser_settings': {
                        'include_node': ['structure', 'trajectory'],
                        'required_node': ['structure', 'trajectory'],
                    }
                },
                extend_list=True,
            )

        # Apply overrides if supplied
        if 'static_calc_settings' in self.inputs:
            self.ctx.static_input_additions.settings = self.inputs.static_calc_settings

        if 'static_calc_options' in self.inputs:
            self.ctx.static_input_additions.options = self.inputs.static_calc_options

        # Override INCARs for the final relaxation
        if 'static_calc_parameters' in self.inputs:
            self.ctx.static_input_additions.parameters = update_nested_dict_node(
                self.inputs.vasp.parameters,
                self.inputs.static_calc_parameters.get_dict(),
            )

        if self.is_verbose():
            if not self.perform_relaxation():
                self.report('Skiped structure relaxation and forwarding inputs to the static calculation.')
            else:
                self.report('performing a final calculation using the relaxed structure.')

        self.ctx.iteration += 1

        inputs = self.exposed_inputs(self._base_workchain, 'vasp')
        inputs.structure = self.ctx.current_structure

        # Attach previous calculation's folder if requested
        if self.ctx.relax_settings.get('reuse', False):
            restart_folder = self.ctx.get('current_restart_folder')  # There might not be any yet
            if restart_folder:
                if self.ctx.get('verbose'):
                    self.report(f'Using previous remote folder <{restart_folder}> for restart')
                inputs.restart_folder = restart_folder

        # The workdir is not cleaned by the called VaspWorkChain for the static calculation
        # if `keep_sp_workdir`` is set to True
        if self.ctx.relax_settings.get('keep_sp_workdir'):
            inputs.keep_last_workdir = orm.Bool(True)

        # Update the MAGMOM if information is present
        if self.ctx.current_magmom is not None:
            inputs.parameters = update_nested_dict_node(
                inputs.parameters,
                {OVERRIDE_NAMESPACE: {'magmom': self.ctx.current_magmom}},
            )

        if 'label' not in inputs.metadata:
            inputs.metadata.label = self.inputs.metadata.get('label', '') + ' SP'

        # Label the calculation and links by iteration number
        inputs.metadata.call_link_label = 'singlepoint'

        # Update the input with whatever stored in the relax_input_additions attribute dict
        inputs.update(self.ctx.static_input_additions)

        # Make sure NSW is not here for the static calculation
        incar = inputs.parameters['incar']
        if 'nsw' in incar:
            incar.pop('nsw')
            new_param = inputs.parameters.get_dict()
            new_param['incar'] = incar
            inputs.parameters = orm.Dict(dict=new_param)
            self.report('Removed explicitly defined NSW value for the static calculation')

        running = self.submit(self._base_workchain, **inputs)
        self.report(f'launching {self._base_workchain.__name__}<{running.pk}> iterations #{self.ctx.iteration}')
        return ToContext(workchains=append_(running))

    def verify_next_workchain(self) -> None | ExitCode:
        """Verify and inherit exit status from child workchains."""

        try:
            workchain = self.ctx.workchains[-1]
        except IndexError:
            self.report(f'There is no {self._base_workchain.__name__} in the called workchain list.')
            return self.exit_codes.ERROR_NO_CALLED_WORKCHAIN  # pylint: disable=no-member

        # Inherit exit status from last workchain (supposed to be
        # successfull)
        next_workchain_exit_status = workchain.exit_status
        next_workchain_exit_message = workchain.exit_message
        if not next_workchain_exit_status:
            self.ctx.exit_code = self.exit_codes.NO_ERROR  # pylint: disable=no-member
        else:
            self.ctx.exit_code = compose_exit_code(next_workchain_exit_status, next_workchain_exit_message)
            self.report(
                'The called {}<{}> returned a non-zero exit status. The exit status {} is inherited'.format(
                    workchain.__class__.__name__, workchain.pk, self.ctx.exit_code
                )
            )

        return self.ctx.exit_code

    def verify_last_relax(self) -> None | ExitCode:
        """Verify and inherit exit status from the last relaxation"""

        try:
            workchain = self.ctx.workchains[-1]
        except IndexError:
            self.report(f'There is no {self._base_workchain.__name__} in the called workchain list.')
            return self.exit_codes.ERROR_NO_CALLED_WORKCHAIN  # pylint: disable=no-member

        # Inherit exit status from last workchain (supposed to be
        # successfull)
        next_workchain_exit_status = workchain.exit_status
        next_workchain_exit_message = workchain.exit_message
        if not next_workchain_exit_status:
            self.ctx.exit_code = self.exit_codes.NO_ERROR  # pylint: disable=no-member
        elif 'misc' in workchain.outputs and 'structure' in workchain.outputs and 'forces' in workchain.outputs.misc:
            self.ctx.exit_code = self.exit_codes.NO_ERROR  # pylint: disable=no-member
            self.report(
                f'The called {workchain.__class__.__name__}<{workchain.pk}> returned a non-zero exit status.'
                ' Continue the workflow as "misc" and "structure" outputs are present.'
            )
        else:
            self.ctx.exit_code = compose_exit_code(next_workchain_exit_status, next_workchain_exit_message)
            self.report(
                'The called {}<{}> returned a non-zero exit status. The exit status {} is inherited'.format(
                    workchain.__class__.__name__, workchain.pk, self.ctx.exit_code
                )
            )

        return self.ctx.exit_code

    def analyze_convergence(self) -> None | ExitCode:
        """
        Analyze the convergence of the relaxation.

        Compare the input and output structures of the most recent relaxation run. If volume,
        shape and ion positions are all within a given threshold, consider the relaxation converged.
        """
        workchain = self.ctx.workchains[-1]
        # Double check presence of structure
        if 'structure' not in workchain.outputs:
            self.report(
                f'The {workchain.__class__.__name__}<{workchain.pk}> for the relaxation run did not have an output '
                'structure and most likely failed. However, its exit status was  zero.'
            )
            return self.exit_codes.ERROR_MISSING_REQUIRED_OUTPUT  # pylint: disable=no-member

        if self.ctx.hybrid_status == 'dft':
            self.report(
                'Competed initial DFT calculation - skipping convergence checks and process to hybrid calculation.'
            )
            self.ctx.hybrid_status = 'hybrid'
            self.ctx.current_restart_folder = workchain.outputs.remote_folder
            return self.exit_codes.NO_ERROR  # pylint: disable=no-member

        # Because the workchain may have been through multiple restarts of the underlying VASP calculation
        # we have to query and find the exact input structure of the calculation that generated the output
        # structure and use that for comparison
        query = orm.QueryBuilder()
        query.append(
            orm.StructureData,
            filters={'id': workchain.outputs.structure.pk},
            tag='workchain-out',
        )
        query.append(orm.CalcJobNode, with_outgoing='workchain-out', tag='calcjob')
        query.append(orm.StructureData, with_outgoing='calcjob')
        input_structure = query.one()[0]

        self.ctx.previous_structure = self.ctx.current_structure
        self.ctx.last_calc_input_structure = input_structure
        self.ctx.current_structure = workchain.outputs.structure

        conv_mode = self.ctx.relax_settings.convergence_mode
        # Assign the two structure used for comparison
        if conv_mode == 'inout':
            compare_from = self.ctx.last_calc_input_structure
            compare_to = self.ctx.current_structure
        elif conv_mode == 'last':
            if 'trajectory' not in workchain.outputs:
                self.report(
                    'Warning - trajectory output not found but needed for convergence check - reverting to check '
                    'input/output structures.'
                )
                compare_from = self.ctx.last_calc_input_structure
                compare_to = self.ctx.current_structure
            else:
                traj = workchain.outputs.trajectory
                if traj.numsteps > 1:
                    compare_from = get_step_structure(
                        workchain.outputs.trajectory, -2
                    )  # take take second last structure
                    compare_to = get_step_structure(workchain.outputs.trajectory, -1)  # take take second last structure
                else:
                    self.report(
                        'Warning - no enough number of steps to compare - using input/output structures instead.'
                    )
                    compare_from = self.ctx.last_calc_input_structure
                    compare_to = self.ctx.current_structure
        else:
            raise RuntimeError(f'Convergence mode {conv_mode} is not a valid option')

        converged = True
        relax_settings = self.ctx.relax_settings
        if relax_settings.convergence_on:
            if self.is_verbose():
                self.report('Checking the convergence of the relaxation.')
            comparison = compare_structures(compare_from, compare_to)
            comparison_inout = compare_structures(self.ctx.last_calc_input_structure, self.ctx.current_structure)
            delta_inout = comparison_inout.absolute if relax_settings.convergence_absolute else comparison.relative
            if relax_settings.positions:
                # For positions it only makes sense to check the absolute change
                converged &= self.check_positions_convergence(comparison.absolute)
            # For volume and shape, always compare the input/output structure
            # This is because VASP only does fixed basis relaxations, so restart is needed if significant check in
            # the cell volume/shape is needed
            if relax_settings.volume:
                converged &= self.check_volume_convergence(delta_inout)
            if relax_settings.shape:
                converged &= self.check_shape_convergence(delta_inout)

            # BONAN: Check force - this is because the underlying VASP calculation may not have finished with
            # fully converge geometry, and the vasp plugin does not check it.
            force_cut_off = relax_settings.get('force_cutoff')
            max_force = get_maximum_force(workchain.outputs.misc.get('forces'))
            if force_cut_off is not None and max_force > force_cut_off:
                self.report(
                    f'Maximum force in the structure {max_force:.4g} excess the cut-off limit {force_cut_off:.4g}'
                    ' - NOT OK'
                )
                converged = False
            elif self.is_verbose():
                self.report(f'Maximum force in the structure {max_force:.4g} - OK')

            if not converged:
                if self.ctx.get('verbose', self._verbose):
                    self.report('Relaxation did not converge, restarting the relaxation.')
            elif self.is_verbose():
                self.report('Relaxation is converged, finishing with a final static calculation.')
        elif self.is_verbose():
            self.report('Convergence checking is not enabled - finishing with a final static calculation.')
        self.ctx.current_restart_folder = workchain.outputs.remote_folder

        # Update the magmom to be used
        if 'site_magnetization' in workchain.outputs and self.ctx.relax_settings.get('keep_magnetization', True):
            try:
                self.ctx.current_magmom = site_magnetization_to_magmom(workchain.outputs.site_magnetization)
            # Some times the site magnetisation can be empty - do nothing
            except ValueError:
                pass

        self.ctx.is_converged = converged

        # Check if we are in 'double_relax' mode and the relaxation is not converged
        if self.ctx.is_double_relax and not converged:
            if self.ctx.iteration >= 2:
                self.report(
                    'DOUBLE RELAX: The second relaxation is not converged, '
                    ' but we will finish with a final static calculation.'
                )
                self.ctx.is_converged = True

        return self.exit_codes.NO_ERROR  # pylint: disable=no-member

    def check_shape_convergence(self, delta: AttributeDict) -> bool:
        """Check the difference in cell shape before / after the last iteratio against a tolerance."""
        threshold_angles = self.ctx.relax_settings.convergence_shape_angles
        threshold_lengths = self.ctx.relax_settings.convergence_shape_lengths

        if threshold_lengths < 0:
            self.report('Cell length convergence check bypassed.')
            lengths_converged = True
        else:
            lengths_converged = bool(delta.cell_lengths.max() <= threshold_lengths)
        if not lengths_converged:
            self.report(
                f'cell lengths changed by max {delta.cell_lengths.max():.4g}, tolerance is {threshold_lengths:.4g}'
                ' - NOT OK'
            )
        elif self.is_verbose():
            self.report(
                f'cell lengths changed by max {delta.cell_lengths.max():.4g}, tolerance is {threshold_lengths:.4g} - OK'
            )

        angles_converged = bool(delta.cell_angles.max() <= threshold_angles)

        if threshold_angles < 0:
            self.report('Cell angles check bypassed.')
            angles_converged = True
        else:
            angles_converged = bool(delta.cell_lengths.max() <= threshold_lengths)

        if not angles_converged:
            self.report(
                f'cell angles changed by max {delta.cell_angles.max():.4g}, tolerance is {threshold_angles:.4g}'
                ' - NOT OK'
            )
        elif self.is_verbose():
            self.report(
                f'cell angles changed by max {delta.cell_angles.max():.4g}, tolerance is {threshold_angles:.4g} - OK'
            )

        return lengths_converged and angles_converged

    def check_volume_convergence(self, delta: AttributeDict) -> bool:
        """Check the convergence of the volume, given a cutoff."""
        threshold = self.ctx.relax_settings.convergence_volume
        if threshold < 0:
            self.report('Volume convergence check bypassed.')
            return True
        volume_converged = bool(delta.volume <= threshold)
        if not volume_converged:
            self.report(f'cell volume changed by {delta.volume:.4g}, tolerance is {threshold:.4g} - NOT OK')
        elif self.is_verbose():
            self.report(f'cell volume changed by {delta.volume:.4g}, tolerance is {threshold:.4g} - OK')

        return volume_converged

    def check_positions_convergence(self, delta: AttributeDict) -> bool:
        """Check the convergence of the atomic positions, given a cutoff."""
        threshold = self.ctx.relax_settings.convergence_positions
        if threshold < 0:
            self.report('Positions convergence check bypassed.')
            return True
        try:
            positions_converged = bool(np.nanmax(delta.pos_lengths) <= threshold)
        except RuntimeWarning:
            # Here we encountered the case of having one atom centered at the origin, so
            # we do not know if it is converged, so settings it to False
            # BONAN: this should never happen now - remove it later
            self.report(
                'there is NaN entries in the relative comparison for the positions during relaxation,'
                ' assuming position is not converged'
            )
            positions_converged = False

        if not positions_converged:
            try:
                self.report(
                    f'max site position change is {np.nanmax(delta.pos_lengths):.4g}, tolerance is {threshold:.4g}'
                    ' - NOT OK'
                )
            except RuntimeWarning:
                pass
        elif self.is_verbose():
            try:
                self.report(
                    f'max site position change is {np.nanmax(delta.pos_lengths):.4g}, tolerance is {threshold:.4g} - OK'
                )
            except RuntimeWarning:
                pass

        return positions_converged

    def store_relaxed(self) -> None | ExitCode:
        """Store the relaxed structure."""
        workchain = self.ctx.workchains[-1]

        relaxed_structure = workchain.outputs.structure
        if self.is_verbose():
            self.report(
                "attaching the node {}<{}> as '{}'".format(
                    relaxed_structure.__class__.__name__,
                    relaxed_structure.pk,
                    'relax.structure',
                )
            )
        self.out('relax.structure', relaxed_structure)

        if not self.ctx.is_converged:
            # If the relaxation is not converged, there is no point to
            # proceed furthure, and the result of last calculation is attached
            workchain = self.ctx.workchains[-1]
            self.out_many(self.exposed_outputs(workchain, self._base_workchain))
            return self.exit_codes.ERROR_RELAX_NOT_CONVERGED  # pylint: disable=no-member

    def results(self) -> None | ExitCode:
        """
        Attach the remaining output results.
        This can either be the final static calculation or the last relaxation if the
        former is not needed.

        As a final check - check if the `maximum_force` is lower than the predefined value.
        """
        workchain = self.ctx.workchains[-1]
        self.out_many(self.exposed_outputs(workchain, self._base_workchain))

        # Try to get the smearing type, there is no point to perform check if the tetrahedral smearing is used.
        if not detect_tetrahedral_method(workchain.inputs.parameters.get_dict()):
            max_force_threshold = self.ctx.relax_settings.get('force_cutoff', 0.03)
            actual_max_force = get_maximum_force(workchain.outputs.misc['forces'])
            if (
                actual_max_force > max(max_force_threshold * 1.5, max_force_threshold + 0.001)
                and self.perform_relaxation()
            ):
                if self.is_verbose():
                    self.report(
                        f'The force of the final SCF is {actual_max_force} eV/A, '
                        'which is significantly higher than the tolerance'
                        f' {max_force_threshold} eV/A.'
                    )
                if not self.ctx.is_double_relax:
                    return self.exit_codes.ERROR_FINAL_SCF_HAS_RESIDUAL_FORCE  # pylint: disable=no-member
        else:
            self.report(
                'Unable to perform final check for maximum force, as the tetrahedral method is used for integration.'
            )

        # If we bypassed the relax by setting perform=False, still attached the input structure as the output structure
        if not RelaxOptions(**self.inputs.relax_settings.get_dict()).perform:
            self.out('relax.structure', workchain.inputs.structure)

    def finalize(self) -> None:
        """
        Finalize the workchain.
        Clean the remote working directories of the called calcjobs
        """
        rlx_settings = self.ctx.relax_settings

        # Fence this section to avoid unnecessary process exceptions.
        try:
            if rlx_settings.get('reuse') and rlx_settings.get('clean_reuse', True):
                self.report('Cleaning remote working directory for the called CalcJobs.')
                cleaned_calcs = []
                qbd = orm.QueryBuilder()
                qbd.append(orm.WorkChainNode, filters={'id': self.node.pk})
                # Options for keeping the single point calculation work directory
                if rlx_settings.get('keep_sp_workdir', False):
                    # Only gather the relax calculations
                    qbd.append(
                        orm.WorkChainNode,
                        edge_filters={'label': {'like': 'relax_%'}},
                        filters={'id': {'in': [node.pk for node in self.ctx.workchains]}},
                    )
                else:
                    qbd.append(
                        orm.WorkChainNode,
                        filters={'id': {'in': [node.pk for node in self.ctx.workchains]}},
                    )

                qbd.append(orm.CalcJobNode)

                # Find the CalcJobs to clean
                if qbd.count() > 0:
                    calcjobs = [tmp[0] for tmp in qbd.all()]
                else:
                    self.report('Cannot found called CalcJobNodes to clean.')
                    return
                # Clean the remote directories one by one
                for calculation in calcjobs:
                    try:
                        calculation.outputs.remote_folder._clean()  # pylint: disable=protected-access
                        cleaned_calcs.append(calculation.pk)
                    except BaseException:  # pylint: disable=no-self-argument,no-self-use
                        pass

                if cleaned_calcs:
                    self.report(f'cleaned remote folders of calculations: {" ".join(map(str, cleaned_calcs))}')
        except BaseException as exception:  # pylint: disable=no-self-argument,no-self-use
            self.report(f'Exception occurred during the cleaning of the remote contents: {exception.args}')

    def perform_relaxation(self) -> bool:
        """Check if a relaxation is to be performed."""
        return self.ctx.relax_settings.perform

    def should_run_static_calculation(self) -> bool:
        """Control whether the static calculation should be run"""
        # If the relaxation itself by passed - run the final static calculation
        if not self.perform_relaxation():
            return True

        # If the shape or volume is relaxed - we must do the final calculation
        relax_settings = self.ctx.relax_settings
        if relax_settings.shape or relax_settings.volume:
            return True
        # Otherwise, we skip the final calculation
        return False

    def is_verbose(self) -> bool:
        """Are we in the verbose mode?"""
        return self.ctx.get('verbose', self._verbose)

    @classproperty
    def relax_option_class(cls) -> RelaxOptions:  # noqa: N805
        """Class for relax options"""
        return RelaxOptions


def compare_structures(structure_a: orm.StructureData, structure_b: orm.StructureData) -> AttributeDict:
    """Compare two StructreData objects A, B and return a delta (A - B) of the relevant properties."""

    delta = AttributeDict()
    delta.absolute = AttributeDict()
    delta.relative = AttributeDict()
    volume_a = structure_a.get_cell_volume()
    volume_b = structure_b.get_cell_volume()
    delta.absolute.volume = np.absolute(volume_a - volume_b)
    delta.relative.volume = np.absolute(volume_a - volume_b) / volume_a

    # Check the change in positions taking account of the pbc
    atoms_a = structure_a.get_ase()
    atoms_a.set_pbc(True)
    atoms_b = structure_b.get_ase()
    atoms_b.set_pbc(True)
    n_at = len(atoms_a)
    atoms_a.extend(atoms_b)
    pos_change_abs = np.zeros((n_at, 3))
    for isite in range(n_at):
        dist = atoms_a.get_distance(isite, isite + n_at, mic=True, vector=True)
        pos_change_abs[isite] = dist

    pos_a = np.array([site.position for site in structure_a.sites])
    # pos_b = np.array([site.position for site in structure_b.sites])
    delta.absolute.pos = pos_change_abs

    site_vectors = [delta.absolute.pos[i, :] for i in range(delta.absolute.pos.shape[0])]
    a_lengths = np.linalg.norm(pos_a, axis=1)
    delta.absolute.pos_lengths = np.array([np.linalg.norm(vector) for vector in site_vectors])
    delta.relative.pos_lengths = np.array([np.linalg.norm(vector) for vector in site_vectors]) / a_lengths

    cell_lengths_a = np.array(structure_a.cell_lengths)
    delta.absolute.cell_lengths = np.absolute(cell_lengths_a - np.array(structure_b.cell_lengths))
    delta.relative.cell_lengths = np.absolute(cell_lengths_a - np.array(structure_b.cell_lengths)) / cell_lengths_a

    cell_angles_a = np.array(structure_a.cell_angles)
    delta.absolute.cell_angles = np.absolute(cell_angles_a - np.array(structure_b.cell_angles))
    delta.relative.cell_angles = np.absolute(cell_angles_a - np.array(structure_b.cell_angles)) / cell_angles_a

    return delta


def get_step_structure(traj: orm.TrajectoryData, step: int) -> orm.StructureData:
    """Get the step structure, but assume the positions are fractional"""
    _, _, cell, symbols, positions, _ = traj.get_step_data(step)
    # Convert to cartesian coorindates
    positions = positions @ cell
    structure = orm.StructureData(cell=cell)
    for _s, _p in zip(symbols, positions):
        # Automatic species generation
        structure.append_atom(symbols=_s, position=_p)
    return structure


def detect_tetrahedral_method(input_dict: dict) -> bool:
    """
    Check if the tetrahedral method is used for BZ integration.
    """
    incar = input_dict.get('incar', {})
    if incar.get('ismear', 1) == -5:
        return True
    if input_dict.get('smearing', {}).get('tetra') and not incar.get('ismear'):
        return True
    return False


class VaspMultiStageRelaxWorkChain(WorkChain, WithBuilderUpdater):
    """
    Relxation with multiple stages

    This workchain allows to run multiple stages of relaxation with different parameters,
    options and settings. The workchain takes a structure as input and runs a series of
    VaspRelaxWorkChain calculations, each with updated set of parameters, options and
    settings as specified in <name>_stages.

    The output of the final workchain is exposed.

    Example::

        vasp_staged_relax = VaspMultiStageRelaxWorkChain.get_builder()
        vasp_staged_relax.structure = structure
        vasp_staged_relax.relax = <usual relax inputs>
        # Set ismear to 0 for the first stage, -5 for the second stage
        vasp_staged_relax.parameters_stages = {
            '0': {'incar': {'ismear': 0}},
            '1': {'incar': {'ismear': 1, 'gga': 'pe', 'lhfcalc': True}},
            }
        # Switch to RMM-DIIS for the second stage
        vasp_staged_relax.relax_settings_stages = {
        '1': {'algo': 'rd'}}
        # Include node in the second stage of the relaxation
        vasp_staged_relax.settings_stages = {
            '1': {'parser_settings': {'inlude_node': ['dos']}},
            }

    Note that the index starts from 0 - e.g. '0' for the first stage, '1' for the second stage, etc.
    """

    _base_workchain = VaspRelaxWorkChain

    @classmethod
    def define(cls, spec: ProcessSpec) -> None:
        super().define(spec)
        spec.input_namespace('parameters_stages', required=False, dynamic=True, help='Parameters for each stage')
        spec.input_namespace('settings_stages', required=False, dynamic=True, help='Settings for each stage')
        spec.input_namespace('options_stages', required=False, dynamic=True, help='Options for each stage')
        spec.input_namespace(
            'relax_settings_stages', required=False, dynamic=True, help='relax_settings for each stage'
        )
        spec.input_namespace(
            'others_stages',
            required=False,
            dynamic=True,
            help='Others for each stage, for example, set "1_kpoints" to apply kpoints to the first stage',
        )
        spec.input(
            'use_nested_update',
            valid_type=orm.Bool,
            default=lambda: orm.Bool(True),
            help='Use nested update for parameters, options, relax_settings, and settings. '
            'Otherwise full dictionary should be provided',
        )
        spec.expose_inputs(cls._base_workchain, 'relax', exclude=('structure',))
        spec.input('structure', valid_type=(orm.StructureData, orm.CifData))
        spec.expose_outputs(cls._base_workchain)

        # Outline of the workchain
        spec.outline(
            cls.setup,
            while_(cls.should_run_stage)(
                cls.run_stage,
                cls.inspect_stage,
            ),
            cls.results,
        )
        spec.exit_code(
            500,
            'ERROR_SUB_PROCESS_FAILED',
            message='The subprocess has failed.',
        )

    def setup(self) -> None:
        """
        Initialize context variables
        """
        relax_inputs = self.exposed_inputs(self._base_workchain, 'relax')
        self.ctx.current_stage = 0
        self.ctx.current_structure = self.inputs.structure
        self.ctx.parameters = relax_inputs.vasp.parameters.get_dict()
        self.ctx.settings = relax_inputs.vasp.settings.get_dict()
        self.ctx.options = relax_inputs.vasp.options.get_dict()
        self.ctx.relax_settings = relax_inputs.relax_settings.get_dict()
        self.ctx.n_stages = len(self.inputs.parameters_stages)

    def should_run_stage(self) -> bool:
        """
        Check if a stage should be run
        """
        return self.ctx.current_stage < self.ctx.n_stages

    def run_stage(self) -> ToContext:
        """
        Run a stage
        """
        self.report(f'Running stage {self.ctx.current_stage}')
        istage = self.ctx.current_stage
        # Update the parameters, options and settings - apply the changes to the self.ctx
        for name in ['parameters', 'options', 'settings', 'relax_settings']:
            if str(istage) in self.inputs.get(name + '_stages', {}):
                if self.inputs.use_nested_update.value:
                    self.ctx[name] = update_nested_dict(
                        self.ctx[name], self.inputs[name + '_stages'][str(istage)].get_dict()
                    )
                else:
                    self.ctx[name] = self.inputs[name + '_stages'][str(istage)].get_dict()

        relax_inputs = self.exposed_inputs(self._base_workchain, 'relax')

        # Apply others changes {'1_relax_settings': {'algo': 'rd'}}
        for key, value in self.inputs.get('others_stages', {}):
            stage = key.split('_')[0]
            name = key.split('_')[1]
            if stage == istage:
                if name in ['static_calc_settings', 'static_calc_options', 'static_calc_parameters']:
                    relax_inputs[name] = value
                else:
                    relax_inputs.vasp[name] = value
        # Modify the inputs
        for name in ['parameters', 'options', 'settings']:
            if relax_inputs.vasp[name].get_dict() != self.ctx[name]:
                relax_inputs.vasp[name] = orm.Dict(dict=self.ctx[name])

        # Modify the relax_settings
        if relax_inputs.relax_settings.get_dict() != self.ctx.relax_settings:
            relax_inputs.relax_settings = orm.Dict(dict=self.ctx.relax_settings)

        relax_inputs.structure = self.ctx.current_structure

        running = self.submit(self._base_workchain, **relax_inputs)
        return ToContext(workchains=append_(running))

    def inspect_stage(self) -> None | ExitCode:
        """
        Inspect a stage
        """
        workchain = self.ctx.workchains[-1]
        self.report(f'Inspecting stage {self.ctx.current_stage} - {workchain}')
        if not workchain.is_finished_ok:
            self.report(f'Stage {self.ctx.current_stage} failed with exit status {workchain.exit_status} - aborting.')
            if not self.inputs.ignored_failed:
                self.out_many(self.exposed_outputs(workchain, self._base_workchain))
                return self.exit_codes.ERROR_SUB_PROCESS_FAILED
            else:
                try:
                    self.ctx.current_structure = workchain.outputs.relax.structure
                except AttributeError:
                    self.report('Cannot get the relaxed structure from the last workchain - aborting.')
                    self.out_many(self.exposed_outputs(workchain, self._base_workchain))
                    return self.exit_codes.ERROR_SUB_PROCESS_FAILED
        else:
            self.ctx.current_structure = workchain.outputs.relax.structure
        self.ctx.current_stage += 1
        return

    def results(self) -> None:
        """
        Attach the remaining output results.
        This can either be the final static calculation or the last relaxation if the
        former is not needed.

        As a final check - check if the `maximum_force` is lower than the predefined value.
        """
        workchain = self.ctx.workchains[-1]
        self.out_many(self.exposed_outputs(workchain, self._base_workchain))


def get_maximum_force(forces: np.ndarray) -> float:
    """Return the maximum value of an array of forces with size (N, 3)"""
    norm = np.linalg.norm(forces, axis=1)
    return np.amax(norm)



"""
Utils for setting symbols on chemical elements.

Utilities for choosing appropriate element symbols for a chemical element.
"""

from __future__ import annotations

import requests
from lxml import html

VERSION = {
    'latest': {
        'version': '5.2',
        'url': 'http://cms.mpi.univie.ac.at/vasp/vasp/Recommended_PAW_potentials_DFT_calculations_using_vasp_5_2.html',
        'gw-url': 'http://cms.mpi.univie.ac.at/vasp/vasp/Recommended_GW_PAW_potentials_vasp_5_2.html',
    }
}


def get_recommendations(version_nr: str = 'latest', use_gw: bool = False) -> dict[str, str]:
    """
    Get recommendations for a certain type of PAW.

    :param version_nr: VASP version number
    :param use_gw: get recommendations for GW instead LDA pseudopotentials
    :return: recommendations dict
    """
    urlkey: str = 'gw-url' if use_gw else 'url'
    page = requests.get(VERSION[version_nr][urlkey], timeout=10)
    tree = html.fromstring(page.text)
    tags = tree.xpath('//table//td[1]/b')
    rec: dict[str, str] = {}
    for tag in tags:
        item: list[str] = tag.text.strip().split(' ')
        element: str = item[0]
        symbol: str = '_'.join([i for i in item if i])
        rec[element] = symbol
    return rec


# pylint: disable=too-few-public-methods
class PawInfo:  # pylint: disable=useless-object-inheritance
    """Simple class to bundle and pass around info about a PAW."""

    def __init__(self, symbol: str, default_enmax: int, valency: float) -> None:
        self.symbol: str = symbol
        self.default_enmax: int = default_enmax
        self.valency: float = valency

    def __str__(self) -> str:
        return self.symbol

    def __repr__(self) -> str:
        return f'<paw: {self.symbol} at {hex(id(self))}>'


def get_all(version_nr: str = 'latest', use_gw: bool = False) -> dict[str, dict[str, PawInfo]]:
    """
    Get recommendations for all symbols.

    :param version_nr: VASP version number
    :param use_gw:  Get recommendations for GW (default: LDA)
    :return: recommendations dict
    """
    urlkey = 'gw-url' if use_gw else 'url'
    page = requests.get(VERSION[version_nr][urlkey], timeout=10)
    tree = html.fromstring(page.text)
    tags = tree.xpath('//table/tr')
    syms = {}
    for tag in tags:
        row = tag.text_content().strip().split('\n')
        if row[1].isdigit():
            symboll = [i for i in row[0].strip().split(' ') if i]
            element = symboll[0]
            symbol = '_'.join(symboll)
            suffix = symboll[1] if len(symboll) > 1 else '_'
            if suffix == 'GW':
                suffix = '_'
            if not syms.get(element):
                syms[element] = {}
            row[0] = symbol
            row[1] = int(row[1])
            row[2] = float(row[2])
            syms[element][suffix] = PawInfo(*row)
    return syms


if __name__ == '__main__':
    DEF_PAW: dict[str, str] = get_recommendations()
    DEF_GW: dict[str, str] = get_recommendations(use_gw=True)
    with open('default_paws.py', 'w', encoding='utf8') as defaults:
        defaults.write('lda = {\n')
        defaults.writelines([f'"{k}": "{v}",\n' for k, v in DEF_PAW.items()])
        defaults.write('}\n\n')
        defaults.write('gw = {\n')
        defaults.writelines([f'"{k}": "{v.replace("_GW", "")}",\n' for k, v in DEF_GW.items()])
        defaults.write('}\n\n')



"""
The ``KPOINTS`` parser interface.

Contains the parsing interfaces to parsevasp used to parse ``KPOINTS`` content.
"""

from __future__ import annotations

from typing import Any, TextIO

import numpy as np
from aiida import orm
from parsevasp.kpoints import Kpoint, Kpoints

from aiida_vasp.parsers.content_parsers.base import BaseFileParser


class KpointsParser(BaseFileParser):
    """The parser interface that enables parsing of ``KPOINTS`` content.

    The parser is triggered by using the ``kpoints-kpoints`` quantity key. The quantity key ``kpoints``
    will on the other hand parse the k-points using the XML parser.

    """

    DEFAULT_SETTINGS = {'quantities_to_parse': ['kpoints-kpoints']}

    PARSABLE_QUANTITIES = {
        'kpoints-kpoints': {
            'inputs': [],
            'name': 'kpoints',
            'prerequisites': [],
        },
    }

    def _init_from_handler(self, handler: TextIO) -> None:
        """Initialize using a file like handler.

        :param handler: A file like object that provides the necessary content to be parsed.
        :type handler: file-like object
        """

        try:
            self._content_parser = Kpoints(file_handler=handler, logger=self._logger)
        except SystemExit:
            self._logger.warning('Parsevasp exited abnormally.')

    def _init_from_data(self, data: orm.KpointsData) -> None:
        """Initialize using AiiDA KpointsData."""

        if isinstance(data, orm.KpointsData):
            self._content_data = data
        else:
            raise TypeError('The supplied AiiDA data structure is not a KpointsData.')

    @property
    def kpoints(self) -> dict[str, Any] | None:
        """Return kpoints that is ready to be consumed by the the AiiDA ``KpointsData``.

        AiiDA does not support the line mode used in VASP, so we give a warning that parsing
        this is not supported.

        :returns: A dict that contain keys ``comment``, ``divisions``, ``shifts``, ``points``, ``tetra``,
                  ``tetra_volume``, ``mode`` ``centering``, ``num_kpoints``, ``weights`` and ``cartesian``
                  which are compatible with consumption of the initialization of the AiiDA KpointsData.
        :rtype: dict
        """

        aiida_kpoints = parsevasp_to_aiida(self._content_parser, self._logger)

        return aiida_kpoints

    def _content_data_to_content_parser(self) -> Any:
        """Convert an AiiDA ``KpointsData`` to a content parser instance of ``Kpoints`` from ``parsevasp``.

        :returns: An instance of ``Kpoints`` from ``parsevasp``.
        :rtype: object
        """
        try:
            # Check if the ``KpointsData`` contain a mesh.
            _ = self._content_data.base.attributes.get('mesh')
            mode = 'automatic'
        except AttributeError:
            pass

        try:
            # Check to see if the ``KpointsData`` contain an explicit k-point list.
            _ = self._content_data.base.attributes.get('array|kpoints')
            mode = 'explicit'
        except AttributeError:
            pass

        kpoints_dict = {}
        for keyword in [
            'comment',
            'divisions',
            'shifts',
            'points',
            'tetra',
            'tetra_volume',
            'mode',
            'centering',
            'num_kpoints',
            'generating_vectors',
        ]:
            kpoints_dict[keyword] = None

        kpoints_dict.update(getattr(self, '_get_kpointsdict_' + mode)(self._content_data))

        # We brake hard if ``parsevasp`` fail here. If we can not write we will not try another parser.
        content_parser = Kpoints(kpoints_dict=kpoints_dict, logger=self._logger)

        return content_parser

    def _get_kpointsdict_explicit(self, kpoints_data: orm.KpointsData) -> dict[str, Any]:
        """Turn Aiida ``KpointData`` into a k-points dictionary with explicit generation of points.

        :param kpoints_data: An AiiDA ``KpointsData`` object containing explicit k-point sets.
        :type kpoints_data: object
        :returns: A dictionary that can be used to initialize a ``parsevasp`` ``Kpoints`` instance.
        :rtype: dict
        """
        kpoints_dict = {}

        kpts = []
        try:
            points, weights = kpoints_data.get_kpoints(also_weights=True)
        except AttributeError:
            points = kpoints_data.get_kpoints()
            weights = None
        for index, point in enumerate(points):
            if weights is not None:
                kpt = Kpoint(point, weight=weights[index])
            else:
                # No weights supplied, so set them to 1.0
                kpt = Kpoint(point, weight=1.0)
            kpts.append(kpt)
        kpoints_dict['points'] = kpts
        kpoints_dict['mode'] = 'explicit'
        kpoints_dict['num_kpoints'] = len(kpts)

        return kpoints_dict

    @staticmethod
    def _get_kpointsdict_automatic(kpointsdata: orm.KpointsData) -> dict[str, Any]:
        """Turn Aiida ``KpointData`` into a k-point dictionary with automatic generation of points.

        :param kpointsdata: An AiiDA ``KpointsData`` object containing meshed k-point sets.
        :type kpointsdata: object
        :returns: A dictionary that can be used to initialize a ``parsevasp`` ``Kpoints`` instance.
        :rtype: dict
        """

        kpoints_dict = {}
        # Automatic mode
        mesh = kpointsdata.get_kpoints_mesh()
        kpoints_dict['divisions'] = mesh[0]
        kpoints_dict['shifts'] = mesh[1]
        kpoints_dict['mode'] = 'automatic'
        # Here we need to make a choice, so should add more to AiiDA to make this better defined
        kpoints_dict['centering'] = 'Gamma'
        kpoints_dict['num_kpoints'] = 0

        return kpoints_dict


def parsevasp_to_aiida(kpoints: Kpoints, logger: Any) -> dict[str, Any] | None:
    """``parsevasp`` to AiiDA conversion.

    Generate an AiiDA data structure that can be consumed by ``KpointsData`` on initialization
    from the ``parsevasp`` instance of the ``Kpoints`` class.

    :param kpoints: An instance of the ``Kpoints`` class in ``parsevasp``.
    :type kpoints: object
    :param logger: Logger instance for warnings and debug messages.
    :type logger: object
    :returns: A dictionary representation which are ready to be used when creating an
              AiiDA ``KpointsData`` instance.
    :rtype: dict
    """

    if kpoints.entries.get('mode') == 'line':
        # AiiDA does not support line mode
        logger.warning('The read KPOINTS contained line mode which isnot supported. Returning None.')
        return None

    # Fetch a dictionary containing the k-points information
    kpoints_dict = kpoints.get_dict()

    # Now unpack points, weights and check direct versus cartesian. Set to
    # None if mode is automatic.
    points = []
    weights = []
    cartesian = []
    for key, value in kpoints_dict.items():
        if key == 'points':
            if value is not None:
                for item in value:
                    points.append(item[0])
                    weights.append(item[1])
                    # AiiDA wants cartesian and not direct flags, so revert
                    cartesian.append(not item[2])
            else:
                points = None
                weights = None
                cartesian = None

    # Make sure weights is ndarray
    if weights is not None:
        weights = np.array(weights)

    # Check that we only have similar elements in the direct list as
    # AiiDA can only work with all points being either in direct or cartesian
    # coordinates.
    if cartesian is not None:
        if not cartesian.count(cartesian[0]) == len(cartesian):
            raise ValueError('Different coordinate systems have been detected among the k-points.')
        cartesian = cartesian[0]

    # Modify dict to AiiDA spec
    kpoints_dict['points'] = points
    kpoints_dict['weights'] = weights
    kpoints_dict['cartesian'] = cartesian

    return kpoints_dict



"""
Fixtures for the VASP calculations.

-----------------------------------
Here we set up different pytest fixtures that are used to represent various VASP
calculations on which one can for instance test parsing etc.
"""

import pathlib

import pytest
from aiida import orm
from aiida.common.extendeddicts import AttributeDict
from aiida.common.folders import SandboxFolder
from aiida.engine.utils import instantiate_process
from aiida.manage.manager import get_manager
from aiida.orm import Dict, List
from aiida.plugins import DataFactory

from aiida_vasp.calcs.base import VaspCalcBase
from aiida_vasp.calcs.neb import VaspNEBCalculation
from aiida_vasp.calcs.vasp import VaspCalculation
from aiida_vasp.calcs.vasp2w90 import Vasp2w90Calculation


@pytest.fixture()
def sandbox_folder():
    """Yield a `SandboxFolder` that can be used for tests where a Folder is needed."""

    with SandboxFolder() as folder:
        yield folder


@pytest.fixture()
def base_calc(aiida_profile, vasp_code):
    """An instance of a VaspCalcBase Process."""

    manager = get_manager()
    runner = manager.get_runner()
    inputs = AttributeDict()

    metadata = AttributeDict({'options': {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}})

    inputs.code = vasp_code
    inputs.metadata = metadata

    return instantiate_process(runner, VaspCalcBase, **inputs)


@pytest.fixture()
def vasp_calc(vasp_inputs):
    """An instance of a VaspCalculation Process."""

    def inner(inputs=None, settings=None):
        if inputs is None:
            inputs = vasp_inputs(settings)
        manager = get_manager()
        runner = manager.get_runner()

        return instantiate_process(runner, VaspCalculation, **inputs)

    return inner


@pytest.fixture()
def vasp_neb_inputs(aiida_profile, vasp_params, vasp_kpoints, vasp_structure, potentials, vasp_code):
    """Inputs dictionary for CalcJob Processes."""

    def inner(settings=None, parameters=None):
        inputs = AttributeDict()

        metadata = AttributeDict({'options': {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}})

        if settings is not None:
            inputs.settings = Dict(dict=settings)

        if isinstance(parameters, dict):
            parameters = orm.Dict(dict=parameters)

        if parameters is None:
            parameters = AttributeDict(vasp_params.get_dict())
            parameters['images'] = 3
            parameters = orm.Dict(dict=parameters)

        inputs.code = vasp_code
        inputs.metadata = metadata
        inputs.parameters = parameters
        inputs.kpoints, _ = vasp_kpoints

        inputs.initial_structure = vasp_structure
        inputs.final_structure = vasp_structure

        inputs.potential = potentials

        neb_images = {f'images_{idx:02d}': vasp_structure for idx in range(1, 4)}
        inputs.neb_images = neb_images

        return inputs

    return inner


@pytest.fixture()
def vasp_neb_calc(vasp_neb_inputs):
    """An instance of a VaspCalculation Process."""

    def inner(inputs=None, settings=None):
        if inputs is None:
            inputs = vasp_neb_inputs(settings)
        manager = get_manager()
        runner = manager.get_runner()

        return instantiate_process(runner, VaspNEBCalculation, **inputs)

    return inner


@pytest.fixture()
def vasp2w90_calc(vasp_inputs):
    """An instance of a VaspCalculation Process."""

    def inner(inputs=None, settings=None):
        if inputs is None:
            inputs = vasp_inputs(settings)

        manager = get_manager()
        runner = manager.get_runner()

        return instantiate_process(runner, Vasp2w90Calculation, **inputs)

    return inner


@pytest.fixture
def vasp_calc_and_ref(vasp_calc, vasp_kpoints, ref_incar):
    """Fixture for non varying setup of a vasp calculation."""
    calc = vasp_calc(settings={'parser_settings': {'add_bands': True, 'add_dos': True}})
    _, ref_kpoints = vasp_kpoints

    return calc, {'kpoints': ref_kpoints, 'incar': ref_incar}


@pytest.fixture
def vasp2w90_calc_and_ref(vasp2w90_calc, vasp_kpoints, vasp2w90_inputs, ref_incar_vasp2w90, ref_win):
    """Fixture for non varying setup of a vasp2w90 calculation."""

    inputs = vasp2w90_inputs(
        settings={
            'parser_settings': {
                'add_bands': True,
                'add_dos': True,
                'poscar_precision': 12,
            }
        }
    )

    calc = vasp2w90_calc(inputs=inputs)
    _, ref_kpoints = vasp_kpoints

    return calc, {'kpoints': ref_kpoints, 'incar': ref_incar_vasp2w90, 'win': ref_win}


@pytest.fixture()
def vasp_chgcar(aiida_profile, data_path):
    """CHGCAR node and reference fixture."""

    chgcar_path = data_path('chgcar', 'CHGCAR')
    chgcar = DataFactory('vasp.chargedensity')(file=chgcar_path)
    with open(chgcar_path, 'r', encoding='utf8') as ref_chgcar_fo:
        ref_chgcar = ref_chgcar_fo.read()
    return chgcar, ref_chgcar


@pytest.fixture()
def vasp_nscf_and_ref(vasp_calc_and_ref, vasp_chgcar, vasp_wavecar):
    """Fixture: vasp calc with chgcar and wavecar given."""
    calc, ref = vasp_calc_and_ref
    chgcar, ref_chgcar = vasp_chgcar
    wavecar, ref_wavecar = vasp_wavecar
    calc.use_charge_density(chgcar)
    calc.use_wavefunctions(wavecar)
    calc.inp.parameters.update_dict({'icharg': 11})
    ref['chgcar'] = ref_chgcar
    ref['wavecar'] = ref_wavecar
    return calc, ref


@pytest.fixture()
def vasp_inputs(aiida_profile, vasp_params, vasp_kpoints, vasp_structure, potentials, vasp_code):
    """Inputs dictionary for CalcJob Processes."""

    def inner(settings=None, parameters=None):
        inputs = AttributeDict()

        metadata = AttributeDict({'options': {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}})

        if settings is not None:
            inputs.settings = Dict(dict=settings)

        if isinstance(parameters, dict):
            parameters = orm.Dict(dict=parameters)

        if parameters is None:
            parameters = AttributeDict(vasp_params.get_dict())
            parameters = orm.Dict(dict=parameters)
        inputs.code = vasp_code
        inputs.metadata = metadata
        inputs.parameters = parameters
        inputs.kpoints, _ = vasp_kpoints
        inputs.structure = vasp_structure
        inputs.potential = potentials

        return inputs

    return inner


@pytest.fixture
def ref_incar(data_path):
    with open(data_path('incar', 'INCAR'), 'r', encoding='utf8') as reference_incar_fo:
        # yield reference_incar_fo.read().strip()
        yield reference_incar_fo.readlines()


@pytest.fixture()
def vasp_wavecar(aiida_profile, data_path):
    """WAVECAR node and reference fixture."""

    wavecar_path = data_path('wavecar', 'WAVECAR')
    wavecar = DataFactory('vasp.wavefun')(file=wavecar_path)
    with open(wavecar_path, 'r', encoding='utf8') as ref_wavecar_fo:
        ref_wavecar = ref_wavecar_fo.read()
    return wavecar, ref_wavecar


@pytest.fixture()
def vasp2w90_inputs(
    aiida_profile,
    vasp_params,
    vasp_kpoints,
    vasp_structure,
    potentials,
    vasp_code,
    wannier_projections,
    wannier_params,
):
    """Inputs dictionary for CalcJob Processes."""

    def inner(settings=None, parameters=None):
        inputs = AttributeDict()

        metadata = AttributeDict({'options': {'resources': {'num_machines': 1, 'num_mpiprocs_per_machine': 1}}})

        if settings is not None:
            inputs.settings = Dict(dict=settings)

        if isinstance(parameters, dict):
            parameters = orm.Dict(dict=parameters)

        if parameters is None:
            parameters = AttributeDict(vasp_params.get_dict())
            parameters = orm.Dict(dict=parameters)

        inputs.code = vasp_code
        inputs.metadata = metadata
        inputs.parameters = parameters
        inputs.kpoints, _ = vasp_kpoints
        inputs.structure = vasp_structure
        inputs.potential = potentials

        inputs.wannier_parameters = wannier_params
        inputs.wannier_projections = wannier_projections

        return inputs

    return inner


@pytest.fixture
def wannier_projections():
    wannier_projections = List()
    wannier_projections.extend(['Ga : s; px; py; pz', 'As : px; py; pz'])
    return wannier_projections


@pytest.fixture
def wannier_params():
    return Dict(
        dict=dict(  # pylint: disable=use-dict-literal
            dis_num_iter=1000,
            num_bands=24,
            num_iter=0,
            num_wann=14,
            spinors=True,
        )
    )


@pytest.fixture
def ref_incar_vasp2w90(data_path):
    with open(data_path('wannier', 'INCAR'), 'r', encoding='utf8') as reference_incar_wannier:
        yield reference_incar_wannier.readlines()


@pytest.fixture
def ref_win(data_path):
    data = pathlib.Path(data_path('wannier90.win'))
    yield data.read_text(encoding='utf8')
